/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2005 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

static const char __idstring[] = "$Id: mx_lz.c,v 1.11 2005/11/07 16:09:04 patrick Exp $";
#include "mx_arch.h"
#include "mx_instance.h"
#include "mx_malloc.h"
#include "mx_misc.h"
#include "mx_peer.h"
#include "mx_pio.h"
#include "mcp_config.h"
#include "mx_stbar.h"
#include "mx_util.h"



/* enable interrupts */
static void
mx_lz_int_enable(mx_instance_state_t *is)
{
  MX_WARN(("%s is just a stub\n", __FUNCTION__));
}

/* disable interrupts */
static void
mx_lz_int_disable(mx_instance_state_t *is)
{
  MX_WARN(("%s is just a stub\n", __FUNCTION__));
}

static void
mx_lz_park_board(mx_instance_state_t *is)
{
  MX_WARN(("%s is just a stub\n", __FUNCTION__));
}

static int
mx_lz_detect_parity(mx_instance_state_t *is)
{
  MX_WARN(("%s is just a stub\n", __FUNCTION__));
  return 0;
}

static int
mx_lz_init_board(mx_instance_state_t *is)
{
  MX_WARN(("FIXME: complete %s \n", __FUNCTION__));
  MX_WARN(("Assuming the MCP was loaded manually with rabbit\n"));

  is->mac_addr_string = "00:01:02:03:04:05";
  is->mac_addr[0] = 0;
  is->mac_addr[1] = 1;
  is->mac_addr[2] = 2;
  is->mac_addr[3] = 3;
  is->mac_addr[4] = 4;
  is->mac_addr[5] = 5;
  return 0;
}

static void
mx_lz_claim_interrupt(mx_instance_state_t *is)
{
  MX_PIO_WRITE((uint32_t *) (is->lanai.sram + 0xfffffc), 0);
}

static int
mx_lz_might_be_interrupting(mx_instance_state_t *is)
{
  if (is->using_msi) {
    return MX_INTR_ACTIVE;
  } else {
    MX_WARN(("%s is not finished, only MSI supported\n", __FUNCTION__));
    return 0;
  }
}

/* map the specials and SRAM.*/

static int
mx_lz_map_board(mx_instance_state_t *is)
{
  /* this is a temporary hack until I know how to interact with
     the mcp/bootloader */

  is->specials_size = 0;
  is->lanai.control_regs = NULL;
  is->lanai.special_regs = NULL;

  /* just pretend we have 2 MB of sram */
  
  MX_INFO(("HACK: Pretending LZ has 2MB SRAM!\n"));

  is->sram_size = 2 * 1024 * 1024;

  is->lanai.sram = mx_map_io_space(is, 0, is->sram_size);
  if (is->lanai.sram == 0)
    return ENODEV;

  return 0;
}

static void
mx_lz_unmap_board(mx_instance_state_t *is)
{
  mx_unmap_io_space(is, is->sram_size, (void*)is->lanai.sram);
  is->lanai.sram = NULL;
}


static void
mx_lz_dump_registers(mx_instance_state_t *is,
		     uint32_t *reg, uint32_t *cnt)
{
   MX_WARN(("%s is just a stub\n", __FUNCTION__));
}

static void
mx_lz_get_freq(mx_instance_state_t *is)
{
  int status;
  
  status = mx_mcpi.get_param(is->id, is->lanai.sram, "clock_freq",
			     &is->cpu_freq);
  
  if (status || is->cpu_freq == 0) {
    MX_WARN(("mx%d: Failed to read LANai clock frequency (%d, %d)\n", 
	     is->id, status, is->cpu_freq));
    return;
  }
  
  /* is->pci_freq = (133.0 * (2 << 12) * 1024) / is->cpu_freq; */
  is->pci_freq = 0;
  return;
}

mx_board_ops_t mx_lz_ops;

void
mx_lz_init_board_ops(void)
{
  mx_lz_ops.init		= mx_lz_init_board;
  mx_lz_ops.map			= mx_lz_map_board;
  mx_lz_ops.unmap		= mx_lz_unmap_board;
  mx_lz_ops.claim_interrupt	= mx_lz_claim_interrupt;
  mx_lz_ops.might_be_interrupting = mx_lz_might_be_interrupting;
  mx_lz_ops.enable_interrupt 	= mx_lz_int_enable;
  mx_lz_ops.disable_interrupt 	= mx_lz_int_disable;
  mx_lz_ops.park		= mx_lz_park_board;
  mx_lz_ops.detect_parity_error	= mx_lz_detect_parity;
  mx_lz_ops.dump_registers	= mx_lz_dump_registers;
  mx_lz_ops.get_freq		= mx_lz_get_freq;
}
