/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2004 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#ifndef _mx_dma_map_h_
#define _mx_dma_map_h_

#include <asm/atomic.h>


#if MX_DMA_DBG
uint8_t *mx_get_lock_count(unsigned long pfn);
void check_bitmap_word(mx_instance_state_t *is, unsigned long addr, int line);
/* debugging the dma-dbg code is disabled by default,
   if something sounds fishy in the dma-dbg code itself
   activate check_bitmap-word by commenting the above line */
#define check_bitmap_word(i,a,l)
#define mx_dma_dbg_lock_kernel() lock_kernel()
#define mx_dma_dbg_unlock_kernel() unlock_kernel()
#else
#define mx_dma_dbg_lock_kernel()
#define mx_dma_dbg_unlock_kernel()
#endif

#define MX_PCI_IOMMU MX_CPU_powerpc64


/* Powerpc64 returns a -1 for error when using pci_map_single (0 might
   be a valid dma address) Every other architecture either uses 0 or
   never fails
*/

#if MX_CPU_powerpc64  
#define INVALID_DMA_ADDR ((dma_addr_t)-1)
#else
#define INVALID_DMA_ADDR ((dma_addr_t)0)
#endif

static inline int
mx_dma_map(mx_instance_state_t *is, struct page *page, mcp_dma_addr_t *dma)
{
  dma_addr_t bus;
#if MX_DMA_DBG
  uint8_t *countp;
  unsigned word, mod;
#endif

  if (!MX_PCI_IOMMU || atomic_read (&is->arch.free_iommu_pages) > 0) {
    bus = pci_map_page (is->arch.pci_dev, page, 0, PAGE_SIZE,
			PCI_DMA_BIDIRECTIONAL);
    if (MX_PCI_IOMMU && bus != INVALID_DMA_ADDR) {
      atomic_dec (&is->arch.free_iommu_pages);
    }
  } else {
    mx_printf_once("Warning: iommu allocation exhausted\n");
    return ENOMEM;
  }
  if (bus == INVALID_DMA_ADDR) {
    mx_printf_once("mx_dma_map:pci_map failure: should be free=%d pages\n",
		   atomic_read(&is->arch.free_iommu_pages));
    return ENXIO;
  } 
  dma->low = MX_LOWPART_TO_U32(bus);
  dma->high = MX_HIGHPART_TO_U32(bus);

#if MX_DMA_DBG
  word = bus - is->lanai.dma_dbg_lowest_addr;
  word /= PAGE_SIZE;
  mx_always_assert(word < mx_dma_pfn_max);
  mod = word & 31;
  word /= 32;

  countp = mx_get_lock_count((bus-is->lanai.dma_dbg_lowest_addr)/PAGE_SIZE);
  check_bitmap_word(is, bus - is->lanai.dma_dbg_lowest_addr, __LINE__);

  if (!*countp) {
    uint32_t val;
    val = is->lanai.dma_dbg_bitmap[word];
    mx_always_assert(!(val & ntohl(1<<mod)));
    val |= ntohl(1<<mod);
    is->lanai.dma_dbg_bitmap[word] = val;
    mb();
  }
  *countp += 1;
  mx_assert(*countp <= 127);
  check_bitmap_word(is, bus - is->lanai.dma_dbg_lowest_addr, __LINE__);
#endif
  return 0;
}

static inline void
mx_dma_unmap (mx_instance_state_t *is, mcp_dma_addr_t *dma)
{
  
  dma_addr_t bus;
#if MX_DMA_DBG
  uint8_t *countp;
  unsigned word, mod;
#endif

  bus = (dma_addr_t)dma->low;
  bus |=  MX_U32_TO_HIGHPART(dma->high, bus);
  
  if (MX_PCI_IOMMU) {
    atomic_inc(&is->arch.free_iommu_pages);
  }
#if MX_DMA_DBG
  countp = mx_get_lock_count((bus-is->lanai.dma_dbg_lowest_addr)/PAGE_SIZE);

  word = bus - is->lanai.dma_dbg_lowest_addr;
  word /= PAGE_SIZE;
  mx_always_assert(word < mx_dma_pfn_max);
  mod = word & 31;
  word /= 32;

  mx_assert(*countp <= 127 && *countp > 0);
  check_bitmap_word(is, bus - is->lanai.dma_dbg_lowest_addr, __LINE__);
  *countp -= 1;
  if (!*countp) {
    uint32_t val;
    val = is->lanai.dma_dbg_bitmap[word];
    mx_always_assert(val & ntohl(1<<mod));
    val &= ~ntohl(1<<mod);
    is->lanai.dma_dbg_bitmap[word] = val;
    mb();
  }
  check_bitmap_word(is, bus - is->lanai.dma_dbg_lowest_addr, __LINE__);
#endif
  pci_unmap_page(is->arch.pci_dev,(dma_addr_t) bus, PAGE_SIZE,
		 PCI_DMA_BIDIRECTIONAL);
}

#endif /* _mx_dma_map_h_ */
