/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2004 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#include <ntddk.h>
#include "mx_auto_config.h"
#include "mx_int.h"
#include "mcp_types.h"
#include "mx_ntddk.h"
#include "mx_arch_io.h"

int 
mx_pin_page(mx_instance_state_t *is, mx_page_pin_t *pin, int flags, uint64_t memory_context)
{
  PMDL mdl;
  PHYSICAL_ADDRESS pa;

  mdl = IoAllocateMdl((void*)(uintptr_t)pin->va, PAGE_SIZE, FALSE,
		      FALSE, NULL);
  if (mdl == NULL) {
    goto abort_with_nothing;
  }
  __try {
    MmProbeAndLockPages(mdl, KernelMode, IoModifyAccess);
    pa = MmGetPhysicalAddress((void*)(uintptr_t)pin->va);
    pin->dma.low = pa.LowPart;
    pin->dma.high = pa.HighPart;
    pin->mdl = mdl;
  }
  __except (EXCEPTION_EXECUTE_HANDLER) {
    goto abort_with_mdl;
  }
  return 0;

 abort_with_mdl:
  IoFreeMdl(mdl);
 abort_with_nothing:
  return EFAULT;
}

void
mx_unpin_page(mx_instance_state_t *is, mx_page_pin_t *pin, int flags)
{
  MmUnlockPages(pin->mdl);
  IoFreeMdl(pin->mdl);
}

#define MX_SCAN_PRIVILEGES 0
#if MX_SCAN_PRIVILEGES
#define MXSTR(x_) #x_
#define MX_CHECK(irp_, x_) \
do { \
  LUID luid; \
  BOOLEAN b; \
  luid = RtlConvertLongToLuid(x_); \
  b = SeSinglePrivilegeCheck(luid, irp_->RequestorMode); \
  DbgPrint(MXSTR(x_) " = %d\n", (int) b); \
} while(0)
#endif

int mx_can_be_privileged(PIRP irp)
{
  LUID lock_memory_luid;
  LUID load_driver_luid;
  LUID impersonate_luid;
  BOOLEAN b;

#if MX_SCAN_PRIVILEGES
  MX_CHECK(irp, SE_MIN_WELL_KNOWN_PRIVILEGE);
  MX_CHECK(irp, SE_CREATE_TOKEN_PRIVILEGE);
  MX_CHECK(irp, SE_ASSIGNPRIMARYTOKEN_PRIVILEGE);
  MX_CHECK(irp, SE_LOCK_MEMORY_PRIVILEGE);
  MX_CHECK(irp, SE_INCREASE_QUOTA_PRIVILEGE);

  MX_CHECK(irp, SE_MACHINE_ACCOUNT_PRIVILEGE);
  MX_CHECK(irp, SE_TCB_PRIVILEGE);
  MX_CHECK(irp, SE_SECURITY_PRIVILEGE);
  MX_CHECK(irp, SE_TAKE_OWNERSHIP_PRIVILEGE);
  MX_CHECK(irp, SE_LOAD_DRIVER_PRIVILEGE);
  MX_CHECK(irp, SE_SYSTEM_PROFILE_PRIVILEGE);
  MX_CHECK(irp, SE_SYSTEMTIME_PRIVILEGE);
  MX_CHECK(irp, SE_PROF_SINGLE_PROCESS_PRIVILEGE);
  MX_CHECK(irp, SE_INC_BASE_PRIORITY_PRIVILEGE);
  MX_CHECK(irp, SE_CREATE_PAGEFILE_PRIVILEGE);
  MX_CHECK(irp, SE_CREATE_PERMANENT_PRIVILEGE);
  MX_CHECK(irp, SE_BACKUP_PRIVILEGE);
  MX_CHECK(irp, SE_RESTORE_PRIVILEGE);
  MX_CHECK(irp, SE_SHUTDOWN_PRIVILEGE);
  MX_CHECK(irp, SE_DEBUG_PRIVILEGE);
  MX_CHECK(irp, SE_AUDIT_PRIVILEGE);
  MX_CHECK(irp, SE_SYSTEM_ENVIRONMENT_PRIVILEGE);
  MX_CHECK(irp, SE_CHANGE_NOTIFY_PRIVILEGE);
  MX_CHECK(irp, SE_REMOTE_SHUTDOWN_PRIVILEGE);
  MX_CHECK(irp, SE_UNDOCK_PRIVILEGE);
  MX_CHECK(irp, SE_SYNC_AGENT_PRIVILEGE);
  MX_CHECK(irp, SE_ENABLE_DELEGATION_PRIVILEGE);
  MX_CHECK(irp, SE_MANAGE_VOLUME_PRIVILEGE);
  MX_CHECK(irp, SE_IMPERSONATE_PRIVILEGE);
  /*MX_CHECK(irp, SE_CREATE_GLOBAL_PRIVILEGE);*/
#endif

  lock_memory_luid = RtlConvertLongToLuid(SE_LOCK_MEMORY_PRIVILEGE);
  load_driver_luid = RtlConvertLongToLuid(SE_LOAD_DRIVER_PRIVILEGE);
  impersonate_luid = RtlConvertLongToLuid(SE_IMPERSONATE_PRIVILEGE);
  b = (SeSinglePrivilegeCheck(lock_memory_luid, irp->RequestorMode) ||
       SeSinglePrivilegeCheck(load_driver_luid, irp->RequestorMode) ||
       SeSinglePrivilegeCheck(impersonate_luid, irp->RequestorMode));
  return (int)b;
}
