/*
 * Structures for dealing with maps and routes
 */

#include <time.h>

/*
 * map info data
 */
struct fma_map_info {
  struct lf_topo_map *current_topo_map;
  int current_topo_map_size;

  uint32_t mi_map_version;
  lf_mac_addr_t mapper_mac_addr;

  int mi_longest_route;		/* longest route from this node to another */
  int mi_most_ports;		/* most ports seen on an xbar */
};

/*
 * Table containing a route to every other NIC
 */
#define FMA_ROUTE_INDEX(NIC, LPORT, RPORT, RT_NUM)	\
	(((NIC) * FMA_FABRIC(A.fabric)->max_nic_ports	\
	  * FMA_FABRIC(A.fabric)->max_nic_ports * 8)	\
      + ((LPORT)* FMA_FABRIC(A.fabric)->max_nic_ports * 8) \
      + ((RPORT) * 8) \
      + (RT_NUM))

#define FMA_ROUTE_LEN(RMT_NIC, LCL_NIC_ID, LCL_IFC, RMT_IFC, RT_NUM) \
	(FMA_NIC(RMT_NIC)->route_lens \
         [(FMA_ROUTE_INDEX(LCL_NIC_ID, LCL_IFC, RMT_IFC, RT_NUM))])

#define FMA_ROUTE_OFFSET(I) ((I) * FMA_IFC_ROUTE_LEN)

struct fma_route_table {
  struct fma_nic_route *nic_routes;	/* one per port on remote NIC */
};
struct fma_nic_route {
  struct fma_ifc_route *ifc_route;	/* one per route */
};
struct fma_ifc_route {
  int route_len;
  unsigned char route[FMA_IFC_ROUTE_LEN];
};


/*
 * prototypes
 */
int fma_copy_and_load_map(struct lf_topo_map *topo_map, int topo_map_size);
void fma_load_map_struct(struct lf_topo_map *topo_map, int topo_map_size);
void fma_release_topo_map(void);
int fma_validate_map(void);
void fma_load_link_state(lf_topo_link_state_t *topo_link_state);
void fma_calc_all_quick_routes(void);
void fma_find_nic_indices(void);
int fma_mac_to_nic_index(lf_mac_addr_t mac);
int fma_set_all_routes(void);
int fma_set_route(struct fma_nic_info *nip, int local_port,
	struct lf_nic *nicp2, int remote_port, int route_num);
int fma_route_topo_map(void);
void fma_start_mapping_fabric(
    void (*map_complete_rtn)(struct lf_topo_map *topo_map, int topo_map_size));
void fma_mf_cancel_mapping(void);
void fma_mf_send_map_to_fms(struct lf_topo_map *topo_map, int topo_map_size);
void fma_standalone_load_map(struct lf_topo_map *topo_map, int topo_map_size);
void fma_standalone_map_is_invalid(void);
struct lf_host *fma_find_host_by_max_mac_addr(lf_mac_addr_t max_mac_addr);
void fma_map_is_invalid(int use_me, char *);
int fma_check_quick_routes(void);
void fma_distribute_topo_map(void);
void fma_set_mapping_level(int level);
void fma_peer_remap_request(struct fma_map_request *req);
