/*
 * Routines for running in standalone mode
 */

#include "fma_myri.h"
#include "fma_myri_packet.h"

#define FMA_STANDALONE_MAP_WAIT_TIMEOUT (60*1000)
#define FMA_FMS_MAP_WAIT_TIMEOUT (300*1000)
#define FMA_MAP_GRACE_PERIOD (20*1000)
#define FMA_MAX_MAP_SENDS_PENDING 50

/*
 * Mapper levels can tell us about other guy's FMS connectivity
 */
#define FMA_MAX_MAPPER_LEVEL 100		/* MAX settable mapper level */
#define FMA_PROXY_MAPPER_LEVEL (FMA_MAX_MAPPER_LEVEL+1)
#define FMA_FMS_MAPPER_LEVEL (FMA_PROXY_MAPPER_LEVEL+1)

/*
 * Data specific to standalone mode
 */
struct fma_standalone_data {
  int active;				/* true if i am the "active" mapper */
  int my_level;

  /* This is true from the time that mapping starts until FMA_MAP_GRACE_PERIOD
   * after map distribution complete
   */
  int mapping_in_progress;
  struct lf_event *map_grace_timer;

  int winner_level;
  lf_mac_addr_t winner_mac;		/* MAC addr of mapper if not me */

  struct lf_host *mapper_host;
  uint8_t mapper_route[MYRI_MAX_ROUTE_LEN];
  int mapper_route_len;
  struct fma_nic_info *mapper_route_nic; /* NIC handle to use */
  int mapper_route_port;

  int map_send_id;
  int map_sends_pending;
  int next_map_send;

  struct lf_event *map_wait_timer;	/* timer for waiting for a map */
  int map_request_msg_id;		/* msg ID of map request in progress */

  /* some static headers we use... */
  struct lf_myri_packet_hdr topo_map_hdr;
};

/*
 * prototypes
 */
void fma_enter_standalone_mode(void);
void fma_cancel_standalone_mode(void);
void fma_exit_standalone_mode(void);
void fma_standalone_got_contact(int level, lf_mac_addr_t mac_addr,
	struct fma_nic_info *nic, int port, uint8_t *route, int route_len,
	int is_lost);
void fma_standalone_map_done(struct lf_topo_map *, int);
void fma_cancel_map_distribution(void);
void fma_reset_mapping_done(void);
