/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2004 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

static const char __idstring[] = "@(#)$Id: mx__endpoint.c,v 1.20 2005/06/29 00:23:15 eugene Exp $";

#include "mx_auto_config.h"
#include "myriexpress.h"
#include "mx__lib_types.h"
#include "mx__request.h"
#  include "mx__partner.h"
#include "mx__endpoint.h"
#include "mx__handle_map.h"

struct mx_endpoint *Mx_endpoints = NULL;

void mx__conservation_of_matter(struct mx_endpoint *ep)
{
  int count = 0;
  union mx_request *r;
  int mcp_handle_count = 0;
#define MX__FOREACH_REQ(r,list) for (r=(list)->basic.next;r!=(list);r=r->basic.next)

  count += mx__count_requests(&ep->send_reqq);
  MX__FOREACH_REQ(r,&ep->send_reqq) {
    switch (r->basic.type) {
    case MX__REQUEST_TYPE_SEND_NULL:
    case MX__REQUEST_TYPE_SEND_SMALL:
    case MX__REQUEST_TYPE_SEND_MEDIUM:
      mx_assert(r->basic.state == MX__REQUEST_STATE_SEND_QUEUED ||
		(r->send.requeued && 
		 (r->basic.state == MX__REQUEST_STATE_DEAD ||
		  r->basic.state == MX__REQUEST_STATE_BUFFERED)));
      break;
    default:
      mx_assert(r->basic.state == MX__REQUEST_STATE_SEND_QUEUED);
    }
  }
  count += mx__count_requests(&ep->recv_reqq);
  MX__FOREACH_REQ(r,&ep->recv_reqq) {
    mx_assert(r->basic.state == MX__REQUEST_STATE_PENDING);
  }
  count += mx__count_requests(&ep->doneq);
  MX__FOREACH_REQ(r,&ep->doneq) {
    mx_assert(r->basic.state & MX__REQUEST_STATE_COMPLETED);
  }
  count += mx__count_requests(&ep->unexpq);
  MX__FOREACH_REQ(r,&ep->unexpq) {
    mx_assert(r->basic.state == MX__REQUEST_STATE_PENDING
	      || r->basic.state == MX__REQUEST_STATE_COMPLETED);
    mx_assert(r->basic.type == MX__REQUEST_TYPE_RECV
	      || r->basic.type == MX__REQUEST_TYPE_RECV_LARGE
	      || r->basic.type == MX__REQUEST_TYPE_RECV_SELF
	      || r->basic.type == MX__REQUEST_TYPE_RECV_SHM);
  }
  count += mx__count_requests(&ep->buffered_sendq);
  mcp_handle_count += mx__count_requests(&ep->buffered_sendq);
  MX__FOREACH_REQ(r,&ep->buffered_sendq) {
    mx_assert(r->basic.state == MX__REQUEST_STATE_DEAD ||
	      r->basic.state == MX__REQUEST_STATE_BUFFERED);
  }
  count += mx__count_requests(&ep->large_sendq);
  MX__FOREACH_REQ(r,&ep->large_sendq) {
    mx_assert(r->basic.state & MX__REQUEST_STATE_MCP);
    mcp_handle_count +=  1;
  }
  count += mx__count_requests(&ep->large_getq);
  mcp_handle_count += mx__count_requests(&ep->large_getq);
  MX__FOREACH_REQ(r,&ep->large_getq) {
    mx_assert(r->basic.state == MX__REQUEST_STATE_MCP);
  }
  count += mx__count_requests(&ep->multifrag_recvq);
  MX__FOREACH_REQ(r,&ep->multifrag_recvq) {
    mx_assert(r->basic.state == MX__REQUEST_STATE_PENDING);
  }
  count += mx__count_requests(&ep->connectq);
  MX__FOREACH_REQ(r,&ep->connectq) {
    mx_assert(!(r->basic.state & ~(MX__REQUEST_STATE_SEND_QUEUED |
				   MX__REQUEST_STATE_MCP |
				   MX__REQUEST_STATE_ACKED |
				   MX__REQUEST_STATE_COMPLETED)));
    if (r->basic.state & MX__REQUEST_STATE_MCP)
      mcp_handle_count += 1;
  }
  count += mx__count_requests(&ep->connect_replyq);
  mcp_handle_count += mx__count_requests(&ep->connect_replyq);
  MX__FOREACH_REQ(r,&ep->connect_replyq) {
    mx_assert(r->basic.state == MX__REQUEST_STATE_MCP);
  }

  mx_always_assert(mx__count_requests(&ep->probeq) == ep->probe_waiters);
  count += ep->probe_count;
  mx_always_assert(mx__count_requests(&ep->peekq) == ep->peek_waiters);
  count += ep->peek_count;

  count += ep->sendself_count;

  count += ep->sendshm_count;

  count += ep->srdma_count;
 
  mcp_handle_count += !!ep->wake_pending;

  mx_always_assert(mcp_handle_count == 
		   ep->handle_map->total_count - ep->handle_map->free_count);
  mx_always_assert(count == ep->req_lookaside.alloc_count - ep->req_lookaside.count);
}

void mx__endpoint_init_eventq(struct mx_endpoint *ep)
{
  ep->eventq_uevt = (volatile mcp_uevt_t*)ep->eventq;
  ep->eventq_index = 0;
  ep->eventq_flow = 0;
  ep->event_count = 0;
}

void mx__endpoint_init_recvq(struct mx_endpoint *ep)
{
  ep->recvq_loc = 0;
}
