/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2004 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#ifndef MX__ENDPOINT_H
#define MX__ENDPOINT_H

#include "mx__request.h"

extern struct mx_endpoint *Mx_endpoints;

static inline union mx_request *
mx__endpoint_match_recv_request(mx_endpoint_t ep, uint64_t match_info, uint32_t msg_length)
{
  union mx_request *r;
  mx_assert(!ep->in_callback);
  r = mx__find_recv_request_by_match_info(&(ep)->recv_reqq, match_info);
  if (!r && ep->callback) {
    ep->in_callback = 1;
    MX__MUTEX_UNLOCK(&ep->lock);
    ep->callback(ep->callback_context, match_info, msg_length);
    MX__MUTEX_LOCK(&ep->lock);
    ep->in_callback = 0;
    /* check if a recv has been posted */
    r = mx__find_recv_request_by_match_info(&(ep)->recv_reqq, match_info);
  }
  return r;
}

#define mx__endpoint_match_partial_request(ep, msg_seq, partner) \
mx__find_partial_recv_request_by_msg_seq(&(partner)->partialq, msg_seq)

#define mx__endpoint_alloc_mcp_handle(ep) mx__hm_alloc((ep)->handle_map)
#define mx__endpoint_free_mcp_handle(ep,idx) mx__hm_free((ep)->handle_map,idx)
#define mx__endpoint_avail_mcp_handle(ep) mx__hm_avail((ep)->handle_map)

static inline
struct mx__partner *
mx__endpoint_lookup_partner(struct mx_endpoint *ep,
			    uint16_t source_endpt,
			    uint16_t source_lookup_hint)
{
  int idx = source_endpt + (source_lookup_hint*ep->max_endpoints);
  if (ep->remote_ep[idx] == NULL) {
    ep->remote_ep[idx] = mx__partner_create(ep, source_lookup_hint, source_endpt);
  }
  return ep->remote_ep[idx];
}

#define mx__endpoint_match_unex(ep, match_info, match_mask) \
mx__find_unex_recv_request_by_matching(&(ep)->unexpq, match_info, match_mask)

void mx__conservation_of_matter(struct mx_endpoint *ep);

void mx__endpoint_init_eventq(struct mx_endpoint *ep);
void mx__endpoint_init_recvq(struct mx_endpoint *ep);

#endif /* ENDPOINT_H */
