/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2004 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

static const char __idstring[] = "@(#)$Id: mx__lib.c,v 1.154.2.7 2006/06/21 03:20:43 loic Exp $";

#include "mx_auto_config.h"
#include "myriexpress.h"
#include "mx__lib_types.h"
#include "mx__lib.h"
#include "mx__fops.h"
#include "mx__lib.h"
#include "mx_util.h"
#include "mx__request.h"
#include "mx__driver_interface.h"
#include "mx__regcache.h"
#include "mx__partner.h"
#include "mx__endpoint.h"
#include "mx__error.h"
#include "mx_byteswap.h"
#include "mx_pin.h"
#include "mx__segment.h"
#include "mx_stbar.h"
#include "mx__handle_map.h"
#include "mx__stack.h"
#include "mx__block.h"
#include "mx__memory_pool.h"
#include "mcp_config.h"
#include "mx__requests.h"
#include "mx__ipconnect.h"
#include "mx__mcp_request_ring.h"
#include "mx__sleep.h"
#include "mx__shmem.h"
#include "mx__shim.h"
#include "mx__internals.h"

/* TODO: Get values from driver instead of header. */

#ifndef MX_KERNEL
uint32_t mx_max_native_endpoints;
uint32_t mx_min_light_endpoints;
uint32_t mx_max_light_endpoints;
#endif

struct mx__opt mx__opt = {0 };

MX_FUNC(mx_return_t)
mx_open_board(int i, mx_endpt_handle_t *handle)
{
  return mx__open(i, -1, handle);
}

MX_FUNC(mx_return_t)
mx_open_any_board(mx_endpt_handle_t *rethandle)
{
  int i;
  mx_endpt_handle_t handle;
  int dev_entry_found;
  int driver_ok;
  int perm_ok;
  mx_return_t ret;

  if (!Mx_init_count)
    return MX_NOT_INITIALIZED;

  /* so far, no good devices, no driver, and no permissions */
  dev_entry_found = 0;
  driver_ok = 0;
  perm_ok = 0;

  /* TODO: Magic number. */
  for (i = 0; i <= 255; ++i) {
    ret = mx__open(i, -1, &handle);

    if (ret == MX_SUCCESS) {
      *rethandle = handle;
      return MX_SUCCESS;

    } else if (ret == MX_NO_DEV) {
      break;
    } else {
      dev_entry_found = 1;

      if (ret == MX_NO_DRIVER) {
	break;
      } else {
        driver_ok = 1;

	if (ret != MX_NO_PERM) {
	  perm_ok = 1;
	}
      }
    }
  }

  if (!dev_entry_found) {
    return MX_NO_DEV;
  }

  if (!driver_ok) {
    return MX_NO_DRIVER;
  }

  if (!perm_ok) {
    return MX_NO_PERM;
  }

  /* default to no resources */
  return MX_NO_RESOURCES;
}


int
mx__queue_large_recv(mx_endpoint_t ep, union mx_request *r)
{
  
  r->recv.basic.status.xfer_length = 
      MX_MIN(r->recv.basic.status.msg_length, r->recv.r_length);
  r->recv.basic.type =  MX__REQUEST_TYPE_RECV_LARGE;
  mx_assert (r->recv.basic.state == MX__REQUEST_STATE_PENDING);
  r->recv.basic.state = MX__REQUEST_STATE_SEND_QUEUED;

  /* FIXME: why can't we try to push the request to the MCP now ? */
  mx__putback_request(&ep->send_reqq, r);
  
  return MX_SUCCESS;

}


static void
mx_unpin(struct mx_endpoint *ep,uint32_t rdma_id,int send)
{
  int cc;
  cc = mx__deregister(ep->handle, rdma_id);
  if (cc != MX_SUCCESS) {
    if (mx_errno == EIO)
      mx_fatal("mx__deregister failed, check kernel logs for error messages");
    else
      mx_always_assert(0);
  }
  mx_rdma_unuse(&ep->rdmas, rdma_id, send);
}


void
mx__regcache_clean(void *ptr, size_t len)
{
  struct mx__rdmawin *r;
  struct mx_endpoint *ep;
  uintptr_t addr = (uintptr_t) ptr;

  if (!mx__opt.rcache)
    return;
  MX__MUTEX_LOCK(&Mx_rcache_lock);
  for (ep = Mx_endpoints; ep; ep = ep->next) {
    struct mx__rdmawin *next;
    for (r = TAILQ_FIRST(&ep->rdmawin_pinned); r; r = next) {
      next = TAILQ_NEXT(r, list);
      if (MX_MIN(r->addr + r->len, addr + len) > MX_MAX(r->addr, addr)) {
	if (mx__opt.verbose) {
	  mx_printf_once("mx__regcache_clean hit:OK\n");
	}
	TAILQ_REMOVE(&ep->rdmawin_pinned, r,list);
	mx_unpin(r->ep, r->rdma_id, r->send);
	TAILQ_INSERT_HEAD(&ep->rdmawin_free,r,list);
      }
    }
  }
  MX__MUTEX_UNLOCK(&Mx_rcache_lock);
}

static void
mx_rdmawin_get(struct mx_endpoint *ep, mx_reg_t *reg_area, int send)
{
  int cc;
  struct mx__rdmawin *r;
  mx_reg_seg_t *seg = reg_area->segs;

  if (mx__opt.rcache) {
    MX__MUTEX_LOCK(&Mx_rcache_lock);
    TAILQ_FOREACH(r, &ep->rdmawin_pinned, list) {
      if (r->send == send
	  && r->addr == seg->vaddr
	  && r->addr + r->len == seg->vaddr + seg->len) {
	break;
      }
    }
    if (r) {
      /* hit shortcut */
      MX__EP_COUNTERS_INC(ep, cache_hit_cnt);
      MX__EP_COUNTERS_ADD(ep, cache_hit_kbytes, seg->len / 1024);
      reg_area->rdma_id = r->rdma_id;
      TAILQ_REMOVE(&ep->rdmawin_pinned, r, list);
      TAILQ_INSERT_HEAD(&ep->rdmawin_free, r, list);
      MX__MUTEX_UNLOCK(&Mx_rcache_lock);
      goto do_register;
    }
    MX__MUTEX_UNLOCK(&Mx_rcache_lock);
    MX__EP_COUNTERS_INC(ep, cache_miss_cnt);
    MX__EP_COUNTERS_ADD(ep, cache_miss_kbytes, reg_area->segs->len / 1024);
  }
  reg_area->rdma_id = mx_rdma_allocate_slot(&ep->rdmas, send);

 do_register:
  mx_assert(reg_area->rdma_id >= 0);
#if MX_OS_UDRV
  if (seg->vaddr < (size_t)mx__init_brk ||
      seg->vaddr > (size_t)mx__init_brk + 256*1024*1024) {
    struct mx_rdma_req *rdma = ep->rdma_requests + reg_area->rdma_id;
    mx_assert(rdma->copy == 0);
    rdma->copy = memalign(MX_RDMA_PAGE_SIZE, seg->len);
    rdma->copy_offset = (size_t)rdma->copy - seg->vaddr;
    if (send)
      memcpy(rdma->copy, (void*)(size_t)seg->vaddr, seg->len);
    seg->vaddr = (size_t)rdma->copy;
  }
#endif
  cc = mx__register(ep->handle, reg_area);
  if (cc != 0) {
    if (mx_errno == EIO) {
      mx_fatal("mx__register failed, check kernel logs for error messages");
    } else {
      MX_WARN(("mx__register_%s: length=%d:%s\n", 
	       send ? "send":"recv", reg_area->segs[0].len,
	       mx_strerrno));
      mx_fatal("mx__register failed");
    }
  }
}

static void 
mx_rdmawin_release(struct mx_endpoint *ep, uint32_t rdma_id, int send, mx_reg_seg_t *seg)
{
  struct mx__rdmawin *r;
  if (mx__opt.rcache) {
    MX__MUTEX_LOCK(&Mx_rcache_lock);
    if (TAILQ_EMPTY(&ep->rdmawin_free)) {
      r = TAILQ_LAST(&ep->rdmawin_pinned, s_rdmawin_pinned);
      TAILQ_REMOVE(&ep->rdmawin_pinned, r, list);
      mx_unpin(r->ep, r->rdma_id, r->send);
    } else {
      r = TAILQ_FIRST(&ep->rdmawin_free);
      TAILQ_REMOVE(&ep->rdmawin_free,r,list);
    }
    r->rdma_id = rdma_id;
    r->send = send;
    r->ep = ep;
    r->addr = (uintptr_t)seg->vaddr;
    r->len = seg->len;
    TAILQ_INSERT_HEAD(&ep->rdmawin_pinned, r, list);
    MX__MUTEX_UNLOCK(&Mx_rcache_lock);
  } else {
    mx_unpin(ep, rdma_id, send);
#if MX_OS_UDRV
    {
      struct mx_rdma_req *rdma;
      rdma = ep->rdma_requests + rdma_id;
      if (rdma->copy) {
	free(rdma->copy);
	rdma->copy = 0;
      }
    }
#endif
  }
}

/* Go through probes looking for one that is satisfied. Return the first
   one that is. */
static union mx_request *
mx__try_probe(mx_endpoint_t ep)
{
  union mx_request *req;

  for (req = ep->probeq.basic.next;
       req != &ep->probeq;
       req = req->basic.next) {
    req->basic.status.context = mx__endpoint_match_unex(ep,
						 req->probe.match_info,
						 req->probe.match_mask);
    if (req->basic.status.context != NULL) {
      break;
    }
  }
  if (req == &ep->probeq) {
    req = NULL;
  }
  return req;
}

void inline
mx__recv_complete(struct mx_endpoint *ep, union mx_request *r, mx_status_code_t status)
{
  r->recv.basic.state = MX__REQUEST_STATE_COMPLETED;
  r->recv.basic.status.xfer_length = 
    MX_MIN(r->recv.basic.status.msg_length, r->recv.r_length);
  if (status == MX_STATUS_SUCCESS) {
    r->recv.basic.status.code = status;
    /* override status code if there was a truncation problem */
    if (r->recv.basic.status.xfer_length < r->recv.basic.status.msg_length) {
      r->recv.basic.status.code = 
	mx__error_req(ep, "recv req", r, MX_STATUS_TRUNCATED);
    }
  } else {
    r->recv.basic.status.code = 
      mx__error_req(ep, "recv req", r, status);
  }
  if (r->basic.wait_state == MX__WAIT_STATE_WAITING) {
    r->basic.wait_state = MX__WAIT_STATE_COMPLETE;
    ep->waiters--;
    MX__EVENT_SIGNAL(&r->basic.event);
  }
  if (ep->peek_waiters) {
    union mx_request *peek_request;
    peek_request = mx__dequeue_request(&ep->peekq);
    peek_request->basic.wait_state = MX__WAIT_STATE_COMPLETE;
    peek_request->basic.status.context = r;
    ep->peek_waiters--;
    MX__EVENT_SIGNAL(&peek_request->basic.event);
  }
  if (r->recv.put_target) {
    mx__rl_free(&ep->req_lookaside, r);
  } else {
    mx__enqueue_request(&ep->doneq, r);
  }
}

#define MX__UNEXP_STORAGE_THRESHOLD (4000000)

static uint32_t
mx__virtual_unexp_length_too_big(mx_endpoint_t ep, uint32_t max)
{
  union mx_request *unexp;
  uint32_t total;
  
  if (ep->ordered_unexp_length > max) {
    return 1;
  }
  total = 0;
  for (unexp=ep->unexpq.basic.next;unexp != &ep->unexpq; unexp = unexp->basic.next) {
    uint32_t weight = 0;
    uint16_t unexp_seq = MX__SEQNO(unexp->recv.msg_seq);
    if (unexp->recv.partner->partialq.basic.part_next != &unexp->recv.partner->partialq) {
      uint16_t first_partial_seq = unexp->recv.partner->partialq.basic.part_next->recv.msg_seq;
      /* (MX__QUADRANT_ONE *2) is the "sign bit" in the SEQNO range */
      if ((unexp_seq - first_partial_seq) & (MX__QUADRANT_ONE * 2)) {
	/* unexp_seq is before first_partial_seq */
	weight = unexp->basic.status.msg_length;
      }
    } else {
      /* no partial */
      weight = unexp->basic.status.msg_length;
    }
    unexp->recv.ordered_unexp_weight = weight;
    total += weight;
  }
  ep->ordered_unexp_length = total;
  if (mx__opt.verbose >= (total > max ? 1 : 2)) {
    mx_printf("unexp_vlength=%d, unexp_length=%d\n", total, ep->unexp_queue_length);
  }
  return (total > max);
}

static union mx_request *
mx__get_req_for_evt(mx_endpoint_t ep, mcp_uevt_recv_t *recv_eager, uint8_t type)
{
  union mx_request *r;
  union mx_request *probe_request;
  uint64_t match_info;
  uint32_t match_a, match_b, msg_length;
  
  match_a = ntohl(recv_eager->match_a);
  match_b = ntohl(recv_eager->match_b);
  match_info = (uint64_t)match_a << 32 | match_b;
  
  if (type == MX_MCP_UEVT_RECV_RNDV)
    /* get the length from the inline data for rdnv messages */
    msg_length = ntohl(*(uint32_t*) (((mcp_uevt_recv_inline_t *) recv_eager)->data));
  else
    msg_length = ntohs(recv_eager->length);

  r = mx__endpoint_match_recv_request(ep, match_info, msg_length);
  if (r == NULL) {
    /* handle unexpected (TODO: A little like mx_irecv) */
    r = mx__rl_alloc(&ep->req_lookaside);
    if (r == NULL) {
      mx_printf_once("INFO: mx__process_recv:mx__rl_alloc failed\n");
      return NULL;
    }
    r->recv.basic.status.match_info = match_info;
    r->recv.segments = &r->recv.segment;
    r->recv.count = 1;
    r->recv.memory_context = MX_PIN_LIBMX_CTX;
    MX__EP_COUNTERS_INC(ep, unexpected);
    r->recv.unexpected = 1;
    r->recv.put_target = 0;
#if MX_ONE_SIDED
    if ((match_info & MX__ONESIDED_MASK) == MX__ONESIDED_PUT) {
      r->recv.segment.segment_ptr = MX_VA_TO_SEGMENT_PTR((void*)(uintptr_t)(match_info & ~MX__ONESIDED_MASK));
      r->recv.segment.segment_length = 0x7fffffff;
      r->recv.unexpected = 0;
      r->recv.put_target = 1;
      r->recv.r_length = 0x7fffffff;
    } else
#endif
    if (type != MX_MCP_UEVT_RECV_RNDV && msg_length) {
      void *segment_ptr;

      if (ep->unexp_queue_length_max != 0
	  && ep->unexp_queue_length > MX__UNEXP_STORAGE_THRESHOLD
	  && mx__virtual_unexp_length_too_big(ep, ep->unexp_queue_length_max)) {
	if (mx__opt.verbose)
	  mx_printf_once("Warning: unexpected queue limit exceeded\n");
	mx__rl_free(&ep->req_lookaside, r);
	return NULL;
      }
      segment_ptr = mx_malloc(msg_length);
      if (segment_ptr == NULL) {
	mx_printf_once("Warning: mx__process_recv/unexp:mx_malloc failed\n");
	mx__rl_free(&ep->req_lookaside, r);
	return NULL;
      }
      r->recv.ordered_unexp_weight = 0;
      ep->unexp_queue_length += msg_length;
      r->recv.segment.segment_ptr = MX_VA_TO_SEGMENT_PTR(segment_ptr);
      r->recv.segment.segment_length = msg_length;
    } else {
      r->recv.segment.segment_ptr = 0;
      r->recv.segment.segment_length = 0;
    }
    r->recv.basic.state = MX__REQUEST_STATE_PENDING;
    r->recv.basic.status.code = MX_STATUS_PENDING;
    if (type == MX_MCP_UEVT_RECV_RNDV) {
      r->recv.basic.type = MX__REQUEST_TYPE_RECV_LARGE;
    }
    else {
      r->recv.basic.type = MX__REQUEST_TYPE_RECV;
    }
    if (!r->recv.put_target)
      mx__enqueue_request(&ep->unexpq, r); 
    if (ep->probe_waiters) {
      probe_request = mx__try_probe(ep);
      if (probe_request != NULL) {
	/* mx__try_probe filled in the context */
	probe_request->basic.wait_state = MX__WAIT_STATE_COMPLETE;
	ep->probe_waiters--;
	mx__spliceout_request(&ep->probeq, probe_request);
	MX__EVENT_SIGNAL(&probe_request->basic.event);
      }
    }
  }
  else {
    mx_assert(r->recv.unexpected == 0);
    /* Matched, take out of recv_reqq. */
    mx__spliceout_request(&ep->recv_reqq, r);
  }
  r->recv.basic.status.msg_length = msg_length;
  r->recv.basic.status.match_info = match_info;
  return r;
}
  

static void
mx__process_recv_null(mx_endpoint_t ep, union mx_request *r, mcp_uevt_recv_t *evt, void *dummy)
{
  r->recv.basic.state = MX__REQUEST_STATE_COMPLETED;
  r->recv.basic.status.code = MX_STATUS_SUCCESS;
  if (r->recv.unexpected == 0) {
    mx__recv_complete(ep, r, MX_STATUS_SUCCESS);
  }
}

static void
mx__process_recv_inline(mx_endpoint_t ep, union mx_request *r, mcp_uevt_recv_t *evt, void *dummy)
{
  mcp_uevt_recv_inline_t *recv_inline = (mcp_uevt_recv_inline_t *)evt;

  mx__copy_to_segments(r->recv.segments, r->recv.count,
		       r->recv.memory_context,
		       0, (char *)recv_inline->data, r->recv.basic.status.msg_length);
  r->recv.basic.state = MX__REQUEST_STATE_COMPLETED;
  r->recv.basic.status.code = MX_STATUS_SUCCESS;
  if (r->recv.unexpected == 0) {
    mx__recv_complete(ep, r, MX_STATUS_SUCCESS);
  } 
#if MX_ONE_SIDED
  else if ((r->recv.basic.status.match_info & MX__ONESIDED_MASK) ==  MX__ONESIDED_GET) {
    mx_segment_t seg;
    mx_request_t req;
    mx_return_t rc;
    mx__get_t *get;

    get = (void*)recv_inline->data;
    seg.segment_ptr = MX_VA_TO_SEGMENT_PTR((void*)(uintptr_t)get->data_target);
    seg.segment_length = get->length;
    ep->in_callback += 1;
    MX__MUTEX_UNLOCK(&ep->lock);
    rc = mx_isend(ep, &seg, 1, r->recv.basic.status.source, 
			r->recv.basic.status.match_info | MX__ONESIDED_MASK, NULL, &req);
    mx_always_assert(rc == MX_SUCCESS);
    MX__MUTEX_LOCK(&ep->lock);
    ep->in_callback -= 1;
    mx__spliceout_request(&ep->unexpq, r);
    mx__rl_free(&ep->req_lookaside, r);
  }
#endif
}

static void
mx__process_recv_copy_single(mx_endpoint_t ep, union mx_request *r, mcp_uevt_recv_t *evt, void *data)
{
  mx__copy_to_segments(r->recv.segments, r->recv.count,
		       r->recv.memory_context,
		       0, data, r->recv.basic.status.msg_length);
  r->recv.basic.state = MX__REQUEST_STATE_COMPLETED;
  r->recv.basic.status.code = MX_STATUS_SUCCESS;
  if (r->recv.unexpected == 0) {
    mx__recv_complete(ep, r, MX_STATUS_SUCCESS);
  }
}

static void
mx__process_recv_copy_frag(mx_endpoint_t ep, union mx_request *r, mcp_uevt_recv_t *evt, void *data, int from_partial)
{
  uint32_t frame_seqnum, frame_length;
  mcp_uevt_recv_copy_multi_t *recv_eager = (mcp_uevt_recv_copy_multi_t *)evt;

  frame_seqnum = recv_eager->frame_seqnum;
  frame_length = ntohs(recv_eager->frame_length);
  
  if (r->recv.r_mask & (1 <<frame_seqnum)) {
    /* duplicate */
    return;
  }
  mx__copy_to_segments(r->recv.segments, r->recv.count,
		       r->recv.memory_context,
		       (1 << recv_eager->pipeline_log)
		       * frame_seqnum, data, frame_length);
  
  r->recv.accum += frame_length;
  r->recv.r_mask |= (1 << frame_seqnum);
  mx_assert(r->recv.accum <= r->recv.basic.status.msg_length);
  if (r->recv.accum == r->recv.basic.status.msg_length) {
    if (from_partial) {
      mx__spliceout_part_request(&r->recv.partner->partialq, r);
    }
    r->recv.basic.state = MX__REQUEST_STATE_COMPLETED;
    r->recv.basic.status.code = MX_STATUS_SUCCESS;
    if (r->recv.unexpected == 0) {
      if (from_partial) {
	mx__spliceout_request(&ep->multifrag_recvq, r);
      }
      mx__recv_complete(ep, r, MX_STATUS_SUCCESS);
    }
  }
}

static void
mx__process_recv_copy_multi(mx_endpoint_t ep, union mx_request *r, mcp_uevt_recv_t *evt, void *data)
{
  mcp_uevt_recv_t *recv_eager = evt;
  r->recv.accum = 0;
  r->recv.r_mask = 0;
  mx__process_recv_copy_frag(ep, r, recv_eager, data, 0);
  if (r->recv.basic.state != MX__REQUEST_STATE_COMPLETED) {
    if (r->recv.unexpected == 0) {
      mx__enqueue_request(&ep->multifrag_recvq, r);
    }
    /* TODO: Insert in order of msg_seqnum? */
    mx__enqueue_part_request(&r->recv.partner->partialq, r);
  }
}


static void
mx__process_recv_large(mx_endpoint_t ep, union mx_request *r, mcp_uevt_recv_t *evt, void *dummy)
{
  mcp_uevt_recv_inline_t *recv_inline = (mcp_uevt_recv_inline_t *)evt;

  /* the following field is only used for big messages.
   * the first 4 bytes have already been used to get the length
   * in mx__get_req_for_evt()
   */
  r->recv.remote_rdma = *(uint32_t *) (recv_inline->data + 4);
  if (!r->recv.unexpected) {
    mx__queue_large_recv(ep, r);
  }
}

static void
mx__obsolete(mx_endpoint_t ep, struct mx__partner *partner, 
	     int order, int msg_seqnum)
{
  mx_always_assert(order < 0);
  if (mx__opt.verbose) {
    if (MX__SESNO(msg_seqnum ^ partner->recv_seq)) {
      mx_printf("mx__proces_recv:received message from previous session\n");
    } else {
      mx_printf_once("INFO:mx__proces_recv:Redundant obsolete message"
		     "(msg=0x%x,partner=0x%x,order=%d)..OK\n",
		     msg_seqnum, partner->recv_seq, order);
    }
  }
}

static int inline
mx__process_ordered_evt(mx_endpoint_t ep, struct mx__partner *partner, int msg_seqnum, 
			int type, mcp_uevt_recv_t *evt, void *data, 
			mx__process_recv_t recv_func)
{
  union mx_request *r;
  mx__endpoint_addr_t *source;
  r = mx__get_req_for_evt(ep, evt, type);
  if (!r)
    return 1;

  source = (mx__endpoint_addr_t*)&r->recv.basic.status.source;
  source->partner = partner;
  source->endpoint_id = evt->source_endpt;
  source->peer_index_n = evt->source_peer_index;
  
  r->recv.partner = partner;
  partner->recv_seq = MX__SESNO(partner->recv_seq) | MX__SEQNO(partner->recv_seq + 1);
  r->recv.msg_seq = msg_seqnum;
  (*recv_func)(ep, r, evt, data);
  return 0;
}



void
mx__process_early(mx_endpoint_t ep, struct mx__partner *partner)
{
  struct mx__early *early, *e;
  int order;
  union mx_request *r;
  mcp_uevt_recv_t *evt;

  early = partner->early.next;
  while (early != &partner->early) {
    order = mx__msg_order(early->msg_seq, partner->recv_seq);
    if (order > 0)
      return;
    e = early;
    early = early->next;
    evt = (mcp_uevt_recv_t*)&e->recv_inline;
    if (order == 0) {
      if (mx__process_ordered_evt(ep, partner, e->msg_seq, e->type, evt, e->data,
				  e->recv_func))
	return;
    } else {
      r = mx__endpoint_match_partial_request(ep, e->msg_seq, partner);
      if (r) {
	/* the only case of late message in the early queue are other
	   fragments of the message just taken into account */
	mx_assert(MX__SEQNO(partner->recv_seq) == MX__SEQNO(e->msg_seq + 1));
	mx_assert(e->type == MX_MCP_UEVT_RECV_COPY_MULTI);
	mx__process_recv_copy_frag(ep, r, evt, e->data, 1);
      } else {
	mx__obsolete(ep, partner, order, e->msg_seq);
      }
    }
    e->prev->next = e->next;
    e->next->prev = e->prev;
    if (e->data) {
      mx_free(e->data);
      e->data = NULL;
    }
    mx_free(e);
  }
}

static int inline
mx__process_recvs(mx_endpoint_t ep, int type, mcp_uevt_recv_t *evt, void *data,
		  mx__process_recv_t recv_func)
{
  int order;
  struct mx__partner *partner;
  union mx_request *r;
  uint32_t msg_seqnum;
  
  msg_seqnum = htons(evt->lib_seqnum);
  partner = mx__endpoint_lookup_partner(ep, evt->source_endpt, htons(evt->source_peer_index));
  order = mx__msg_order(msg_seqnum, partner->recv_seq);
  if (order == 0) {
    if (mx__process_ordered_evt(ep, partner, msg_seqnum, type, evt, data, recv_func))
      return 1;
    mx__process_early(ep, partner);
    return 0;
  } else if (order < 0 && data && 
	     (r = mx__endpoint_match_partial_request(ep, msg_seqnum, partner))) {
    mx_assert(type == MX_MCP_UEVT_RECV_COPY_MULTI);
    mx__process_recv_copy_frag(ep, r, evt, data, 1);
    return 0;
  } else  if (order > 0) {
    if (ep->unexp_queue_length_max && partner->early.next &&
	htons(partner->early.next->recv_inline.lib_seqnum) == partner->recv_seq) {
      /* don't insert early if we stopped processing the early list */
      mx_assert(mx__virtual_unexp_length_too_big(ep, ep->unexp_queue_length_max));
      return 1;
    }
    mx_assert(!partner->early.next || 
	      htons(partner->early.next->recv_inline.lib_seqnum) != partner->recv_seq);
      
    MX__EP_COUNTERS_INC(ep, early);
    if (mx__partner_insert_early(partner, evt,
				 msg_seqnum, 
				 recv_func,
				 type, data) == NULL) {
      mx_fatal("mx__partner_insert_early failed");
      return 1;
    }
    return 0;
  } else {
    mx__obsolete(ep, partner, order, msg_seqnum);
    return 0;
  }
}


void inline
mx__send_complete(mx_endpoint_t ep, union mx_request *r, mx_status_code_t status)
{
  union mx_request *peek_request;

  if (r->send.count > 1) {
    mx_free(r->send.segments);
  }
  if (!(r->send.basic.state & MX__REQUEST_STATE_DEAD)) {
    /* The request is complete, but the user hasn't tested or waited
       on it yet. */
    mx_assert(!(r->send.basic.state & (MX__REQUEST_STATE_MCP |
				       MX__REQUEST_STATE_SEND_QUEUED)));
    r->send.basic.state = MX__REQUEST_STATE_COMPLETED;
    r->send.basic.status.msg_length = r->send.length;
    /* TODO: Use generic status translation function/macro. */
    if (status == MX_STATUS_SUCCESS) {
      r->send.basic.status.code = MX_STATUS_SUCCESS;
      if (r->send.basic.status.xfer_length < r->send.length) {
	r->send.basic.status.code = 
	  mx__error_req(ep, "send req", r, MX_STATUS_TRUNCATED);
      }
    }
    else {
      r->send.basic.status.code = 
	mx__error_req(ep, "send req", r, status);
    }
#if MX_ONE_SIDED
    if (r->send.match_info & MX__ONESIDED_GET) {
      mx__get_t *get = r->basic.status.context;
      if (get) {
	mx_free(get);
      }
      mx__rl_free(&ep->req_lookaside, r);
      return;
    }
#endif
    mx__enqueue_request(&ep->doneq, r);
    if (r->basic.wait_state == MX__WAIT_STATE_WAITING) {
      r->basic.wait_state = MX__WAIT_STATE_COMPLETE;
      ep->waiters--;
      MX__EVENT_SIGNAL(&r->basic.event);
    }
    if (ep->peek_waiters) {
      peek_request = mx__dequeue_request(&ep->peekq);
      peek_request->basic.wait_state = MX__WAIT_STATE_COMPLETE;
      peek_request->basic.status.context = r;
      ep->peek_waiters--;
      MX__EVENT_SIGNAL(&peek_request->basic.event);
    }
  }
  else {
    /* Someone already tested or waited on handle. The request is
       complete. */
    if (status != MX_STATUS_SUCCESS) {
      if (!ep->cancelled) {
	/* only handle errors when the endpoint is totally open */
	mx__error_req(ep, "send req(already completed)", r, status);
      }
    }
    mx__rl_free(&ep->req_lookaside, r);
  }
}

static void
mx__finish_recv_large(struct mx_endpoint *ep, union mx_request *r, 
			   mx_status_code_t status)
{
  mx__spliceout_request(&ep->large_getq, r);
  mx__recv_complete(ep, r, status);
#if MX_OS_UDRV
 {
   struct mx_rdma_req *rdma;
   rdma = ep->rdma_requests + r->recv.local_rdma_id;
   if (rdma->copy) {
     memcpy((char*)r->recv.segment.segment_ptr, 
	    (char*)r->recv.segment.segment_ptr + rdma->copy_offset,
	    r->recv.basic.status.xfer_length);
   }
  }
#endif
  mx_rdmawin_release(ep, r->recv.local_rdma_id, 0, 
		     r->recv.reg_area.segs);
  ep->rdma_requests[r->recv.local_rdma_id].req = 0;
}

static int
mx__process_events(struct mx_endpoint * ep, unsigned int event_count)
{
  unsigned int i, recvq_offset;
  uint32_t rdma_id;
  union mx_request *r;
  /* TODO: Should mcp_event be volatile? */
  /* EVENT: uevt_recv_eager */
  mcp_uevt_t *mcp_event;
  mx_status_code_t status;
  uint8_t type;
  uint16_t mcp_handle;

  i = 0; 
  do {
    mcp_event = (mcp_uevt_t *) ep->eventq_uevt;
    type = mcp_event->basic.type;
    if (type == 0)
      return 0;
    MX_READBAR();
    switch(type) {
    case MX_MCP_UEVT_DONE_SUCCESS:
      mcp_handle =  ntohs(mcp_event->done.uevt.lib_cookie);
      r = mx__hm_get(ep->handle_map, mcp_handle);
      /* Return resources. Buffered sends no longer have to be waited on
         at endpoint close. */
      switch (r->send.basic.type) {
      case MX__REQUEST_TYPE_SEND_NULL:
	mx_assert(r->basic.state == MX__REQUEST_STATE_BUFFERED ||
		  r->basic.state == MX__REQUEST_STATE_DEAD);
	/* Take out of buffered_senq. */
	mx__spliceout_request(&ep->buffered_sendq, r);
	break;
      case MX__REQUEST_TYPE_SEND_SMALL:
	mx_assert(r->basic.state == MX__REQUEST_STATE_BUFFERED ||
		  r->basic.state == MX__REQUEST_STATE_DEAD);
	if (r->send.length > 0) {
	  mx__ptr_stack_push(ep->small_msg_ptrs,
			     ep->udataq + r->send.offset_data);
	  mx__ptr_stack_push(ep->shadow_small_msg_ptrs,
			     ep->shadow_udataq + r->send.offset_data);
	}
	/* Take out of buffered_senq. */
	mx__spliceout_request(&ep->buffered_sendq, r);
	break;
      case MX__REQUEST_TYPE_SEND_MEDIUM:
	mx_assert(r->basic.state == MX__REQUEST_STATE_BUFFERED ||
		  r->basic.state == MX__REQUEST_STATE_DEAD);
	mx__memory_pool_free(ep->send_pool, ep->sendq 
			     + (r->send.offset_data << MX_MCP_VPAGE_SHIFT),
			     ep->medium_msg_threshold);
	/* Take out of buffered_sendq. */
	mx__spliceout_request(&ep->buffered_sendq, r);
	break;
      case MX__REQUEST_TYPE_SEND_LARGE:
	mx_assert(r->basic.state & MX__REQUEST_STATE_MCP);
	r->basic.state &= ~MX__REQUEST_STATE_MCP;
	r->basic.state |= MX__REQUEST_STATE_ACKED;
	mx__spliceout_request(&ep->large_sendq, r);
	if (!(r->basic.state & MX__REQUEST_STATE_NOTIFIED))
	  ep->srdma_count += 1;
	break;
      case MX__REQUEST_TYPE_CONNECT:
	mx_assert(r->basic.state & MX__REQUEST_STATE_MCP);
	r->basic.state &= ~MX__REQUEST_STATE_MCP;
	r->basic.state |= MX__REQUEST_STATE_ACKED;
	mx__wake_request(ep, r);
	goto finish_send_processing;
      case MX__REQUEST_TYPE_CONNECT_REPLY:
	mx_assert(r->basic.state == MX__REQUEST_STATE_MCP);
	mx__spliceout_request(&ep->connect_replyq, r);
	mx__rl_free(&ep->req_lookaside, r);
	goto finish_send_processing;
      case MX__REQUEST_TYPE_RECV_LARGE:
	mx_assert(r->basic.state == MX__REQUEST_STATE_MCP);
	mx__finish_recv_large(ep, r, MX_STATUS_SUCCESS);
	goto finish_send_processing;
      default:
	MX_PRINT(("unknown send type %d\n", (int)r->send.basic.type));
	mx_assert(0);
	break;
      }
      
      mx_assert(r->send.partner->quadrant_count[MX__QUADRANT(r->send.send_seq)] > 0);
      r->send.partner->quadrant_count[MX__QUADRANT(r->send.send_seq)] -= 1;
      if (r->send.basic.type != MX__REQUEST_TYPE_SEND_LARGE || 
	  (r->send.basic.state & MX__REQUEST_STATE_NOTIFIED)) {
	/* either less than large completion or
	   the uncommon case of large send completion with scout-ack
	   arriving after the notification
	*/
	mx__send_complete(ep, r, MX_STATUS_SUCCESS);
      }
      
    finish_send_processing:
      mx__endpoint_free_mcp_handle(ep, mcp_handle);
      break;
      
    case MX_MCP_UEVT_DONE_ERROR:
      status = mcp_event->done.uevt.status;
      mcp_handle =  ntohs(mcp_event->done.uevt.lib_cookie);
      r = mx__hm_get(ep->handle_map, mcp_handle);

      if (status == MX_STATUS_EVENTQ_FULL) {

	switch (r->send.basic.type) {
	case MX__REQUEST_TYPE_SEND_NULL:
	case MX__REQUEST_TYPE_SEND_SMALL:
	case MX__REQUEST_TYPE_SEND_MEDIUM:
	  mx_assert(r->basic.state == MX__REQUEST_STATE_BUFFERED ||
		    r->basic.state == MX__REQUEST_STATE_DEAD);
	  mx__spliceout_request(&ep->buffered_sendq, r);
	  r->send.requeued = 1;
	  break;
	case MX__REQUEST_TYPE_SEND_LARGE:
	  mx_assert(r->basic.state & MX__REQUEST_STATE_MCP); 
	  r->basic.state &= ~MX__REQUEST_STATE_MCP;
	  r->basic.state |= MX__REQUEST_STATE_SEND_QUEUED; 
	  mx__spliceout_request(&ep->large_sendq, r);
	  r->send.requeued = 1;
	  break;
	case MX__REQUEST_TYPE_CONNECT:
	  mx_assert(r->basic.state & MX__REQUEST_STATE_MCP);
	  r->basic.state &= ~MX__REQUEST_STATE_MCP;
	  r->basic.state |= MX__REQUEST_STATE_SEND_QUEUED; 
	  mx__spliceout_request(&ep->connectq, r);
	  break;
	case MX__REQUEST_TYPE_CONNECT_REPLY:
	  mx_assert(r->basic.state & MX__REQUEST_STATE_MCP);
	  r->basic.state &= ~MX__REQUEST_STATE_MCP;
	  r->basic.state |= MX__REQUEST_STATE_SEND_QUEUED; 
	  mx__spliceout_request(&ep->connect_replyq, r);
	  break;
	case MX__REQUEST_TYPE_RECV_LARGE:
	  mx_assert(r->basic.state == MX__REQUEST_STATE_MCP); 
	  r->basic.state = MX__REQUEST_STATE_SEND_QUEUED; 
	  mx__spliceout_request(&ep->large_getq, r);
	  mx_rdmawin_release(ep, r->recv.local_rdma_id, 0, 
			     r->recv.reg_area.segs);
	  ep->rdma_requests[r->recv.local_rdma_id].req = 0;
	  break;
	default:
	  MX_PRINT(("unknown send type %d\n", (int)r->send.basic.type));
	  mx_assert(0);
	  break;
	}
	
	mx__putback_request(&ep->send_reqq, r);
	mx__endpoint_free_mcp_handle(ep, mcp_handle);
      } else {
	/* Return resources. Buffered sends no longer have to be waited on
	   at endpoint close. */
	switch (r->send.basic.type) {
	case MX__REQUEST_TYPE_SEND_NULL:
	  mx_assert(r->basic.state == MX__REQUEST_STATE_BUFFERED ||
		    r->basic.state == MX__REQUEST_STATE_DEAD);
	  /* Take out of buffered_senq. */
	  mx__spliceout_request(&ep->buffered_sendq, r);
	  break;
	case MX__REQUEST_TYPE_SEND_SMALL:
	  mx_assert(r->basic.state == MX__REQUEST_STATE_BUFFERED ||
		    r->basic.state == MX__REQUEST_STATE_DEAD);
	  if (r->send.length > 0) {
	    mx__ptr_stack_push(ep->small_msg_ptrs,
			       ep->udataq + r->send.offset_data);
	    mx__ptr_stack_push(ep->shadow_small_msg_ptrs,
			       ep->shadow_udataq + r->send.offset_data);
	  }
	  /* Take out of buffered_senq. */
	  mx__spliceout_request(&ep->buffered_sendq, r);
	  break;
	case MX__REQUEST_TYPE_SEND_MEDIUM:
	  mx_assert(r->basic.state == MX__REQUEST_STATE_BUFFERED ||
		    r->basic.state == MX__REQUEST_STATE_DEAD);
	  mx__memory_pool_free(ep->send_pool, ep->sendq 
			       + (r->send.offset_data << MX_MCP_VPAGE_SHIFT),
			       ep->medium_msg_threshold);
	  /* Take out of buffered_sendq. */
	  mx__spliceout_request(&ep->buffered_sendq, r);
	  break;
	case MX__REQUEST_TYPE_SEND_LARGE:
	  mx_assert(r->send.basic.state & MX__REQUEST_STATE_MCP);
	  r->send.basic.state &= ~MX__REQUEST_STATE_MCP;
	  r->send.basic.state |= MX__REQUEST_STATE_ACKED;
	  rdma_id = r->send.local_rdma_id;
	  mx__spliceout_request(&ep->large_sendq, r);
	  mx_rdmawin_release(ep, rdma_id, 1, r->send.reg_area.segs);
	  ep->rdma_requests[rdma_id].req = 0;
	  break;
	case MX__REQUEST_TYPE_CONNECT:
	  mx_assert(r->basic.state & MX__REQUEST_STATE_MCP); 
	  r->basic.state &= ~MX__REQUEST_STATE_MCP;
	  r->basic.state |= MX__REQUEST_STATE_ACKED;
	  r->connect.basic.status.code = 
	    mx__error_req(ep, "connect req", r, status);
	  r->connect.basic.state |= MX__REQUEST_STATE_COMPLETED;
	  mx__wake_request(ep, r);
	  goto finish_error_processing;
	case MX__REQUEST_TYPE_CONNECT_REPLY:
	  mx_assert(r->basic.state == MX__REQUEST_STATE_MCP); 
	  mx__error_req(ep, "connect auto-reply", r, status);
	  mx__spliceout_request(&ep->connect_replyq, r);
	  mx__rl_free(&ep->req_lookaside, r);
	  goto finish_error_processing;
	case MX__REQUEST_TYPE_RECV_LARGE:
	  mx_assert(r->basic.state == MX__REQUEST_STATE_MCP); 
	  mx__finish_recv_large(ep, r, status);
	  goto finish_error_processing;
	default:
	  MX_PRINT(("unknown send type %d\n", (int)r->send.basic.type));
	  mx_assert(0);
	  break;
	}
	
	/* code common to all "real" send requests */
	mx_assert(r->send.partner->quadrant_count[MX__QUADRANT(r->send.send_seq)] > 0);
	r->send.partner->quadrant_count[MX__QUADRANT(r->send.send_seq)] -= 1;
	mx__send_complete(ep, r, status);
      
      finish_error_processing:
	mx__endpoint_free_mcp_handle(ep, mcp_handle);

	/* 
	   LOIC: that was the processing of UEVT_ERROR before we move the 
	   EVENTQ_FULL in UEVT_ERROR. A little harsh, no ? There are 
	   non-fatal errors reported in UEVT_ERROR sometimes...
	
	   mx_printf("MCP error %d for request type %d:%d\n", 
	             mcp_event->error.uevt.status, r->basic.type,
	   mx__dump_request(ep, r);
	   mx_fatal("Receveid MCP_UEVT_ERROR");
	*/
      }
      break;
      
    case MX_MCP_UEVT_CONNECT:
      if (!mcp_event->connect.uevt.is_reply) {
	mx__handle_connect(ep, &mcp_event->connect.uevt);
      } else {
	mx__handle_connect_reply(ep, &mcp_event->connect.uevt);
      }
      break;

    case MX_MCP_UEVT_RECV_NULL:
      if (mx__process_recvs(ep, type, (mcp_uevt_recv_t*)&(mcp_event->recv_copy_single.uevt),
			    0, mx__process_recv_null) != 0) {
	return 1;
      }
      break;
      
    case MX_MCP_UEVT_RECV_INLINE:
      if (mx__process_recvs(ep, type, (mcp_uevt_recv_t*)&(mcp_event->recv_inline.uevt),
			    0, mx__process_recv_inline) != 0) {
	return 1;
      }
      break;
    
    case MX_MCP_UEVT_RECV_COPY_SINGLE:
      recvq_offset = (ntohs(mcp_event->recv_copy_single.uevt.recvq_vpage_index) 
		      << MX_MCP_VPAGE_SHIFT);
      ep->recvq_loc = recvq_offset;
      if (mx__process_recvs(ep, type, (mcp_uevt_recv_t*)&(mcp_event->recv_copy_single.uevt),
			    ep->recvq + recvq_offset, mx__process_recv_copy_single) != 0) {
	return 1;
      }
      break;

    case MX_MCP_UEVT_RECV_COPY_MULTI:
      recvq_offset = (ntohs(mcp_event->recv_copy_multi.uevt.recvq_vpage_index) 
		      << MX_MCP_VPAGE_SHIFT);
      ep->recvq_loc = recvq_offset;
      if (mx__process_recvs(ep, type, (mcp_uevt_recv_t*)&(mcp_event->recv_copy_multi.uevt), 
			    ep->recvq + recvq_offset, mx__process_recv_copy_multi) != 0) {
	return 1;
      }
      break;
      
    case MX_MCP_UEVT_RECV_RNDV:
      if (mx__process_recvs(ep, type, (mcp_uevt_recv_t*)&(mcp_event->recv_inline.uevt), 
			    0, mx__process_recv_large) != 0) {
	return 1;
      }
      break;
      
    case MX_MCP_UEVT_RDMA_NOTIFY:
      rdma_id = mcp_event->rdma_notify.uevt.rdmawin_id;
      mx_assert(rdma_id < (unsigned)ep->rdmas.max);
      r = ep->rdma_requests[rdma_id].req;
      mx_assert(r);
      mx_assert((r->basic.state & (MX__REQUEST_STATE_ACKED | MX__REQUEST_STATE_MCP)) ||
		r->send.requeued);
      mx_assert(r->basic.type == MX__REQUEST_TYPE_SEND_LARGE);
      r->basic.state |= MX__REQUEST_STATE_NOTIFIED;
      r->send.basic.status.xfer_length = ntohl(mcp_event->rdma_notify.uevt.length);
      if (r->basic.state & MX__REQUEST_STATE_ACKED) {
	/* common case of large send completion, notif follows scout-ack */
	mx_assert(ep->srdma_count >= 0);
	ep->srdma_count -= 1;
	mx__send_complete (ep, r, MX_STATUS_SUCCESS);
      }
      mx_assert(mcp_event->rdma_notify.uevt.rdmawin_seqnum == ep->rdma_requests[rdma_id].seqno);
      ep->rdma_requests[rdma_id].seqno += 1;
      mx_rdmawin_release(ep, rdma_id, 1, r->send.reg_area.segs);
      ep->rdma_requests[rdma_id].req = 0;
      break;
      
    default:
      mx_printf("Unknown event type from MCP:%d\n", type);
      mx_fatal("Bailing out");
      break;
    }

    mcp_event->basic.type = 0;
    ep->eventq_uevt++;
    ep->eventq_index++;
    ep->eventq_flow++;
    ep->event_count++;
    i++;
    
    if (ep->eventq_index == (ep->eventq_length / sizeof (mcp_uevt_t))) {
      ep->eventq_uevt = (volatile mcp_uevt_t*)ep->eventq;
      ep->eventq_index = 0;
    }
  } while (i < event_count);
  return 0;
}

static void
mx__post_large_recv(struct mx_endpoint *ep, union mx_request *r, 
		    mcp_ureq_t *req, uint16_t mcp_handle, 
		    uint8_t origin_rdmawin_id, uint8_t origin_rdmawin_seqnum, 
		    uint16_t origin_rdma_offset)
{
  mx__endpoint_addr_t *addr;
  struct mx__partner *partner;
  
  r->connect.basic.state = MX__REQUEST_STATE_MCP;
  addr = (mx__endpoint_addr_t *) &r->recv.basic.status.source;
  partner = mx__endpoint_lookup_partner(ep, addr->endpoint_id, 
					htons(addr->peer_index_n));
  
  /* post the request */
  mx__post_ureq_pull(req, addr, partner, r, origin_rdmawin_id, 
		     origin_rdmawin_seqnum, origin_rdma_offset, mcp_handle);
}


void
mx__post_send(struct mx_endpoint *ep, union mx_request *q,
	      mcp_ureq_t *ureq, uint16_t mcp_handle)
{
  uint32_t length, offset;
  char *p;
  char *shadow;
  uint16_t eid;
  uint16_t peer;
  struct mx__partner *partner;
  uint32_t i, chunk;
  uint16_t send_seq;

  q->send.basic.mcp_handle = mcp_handle;
  /* TODO: Encapsulate in partner.c? */
  eid = q->send.dest_address.endpoint_id;
  peer = htons(q->send.dest_address.peer_index_n);
  partner = mx__endpoint_lookup_partner(ep, eid, peer);

  /* TODO: Optimize for non requeued case. */
  if (q->send.requeued) {
    send_seq = q->send.send_seq;
  }
  else {
    send_seq = partner->send_seq;
    q->send.send_seq = partner->send_seq;
    q->send.partner = partner;
    mx_assert(partner->quadrant_count[MX__QUADRANT(send_seq)] < ep->max_mcp_handles);
    partner->quadrant_count[MX__QUADRANT(send_seq)] += 1;
    partner->send_seq = MX__SESNO(send_seq) | MX__SEQNO(send_seq + 1);
  }

  /* TODO: mx__post_ureq_* should only be called when nothing can go wrong. */
  length = q->send.length;
  switch (q->send.basic.type) {
  case MX__REQUEST_TYPE_SEND_NULL:
    if (q->send.requeued) {
      mx__post_ureq_null(ureq, &q->send.dest_address, partner, 
			 q->send.match_info, send_seq, mcp_handle);
      mx__enqueue_request(&ep->buffered_sendq, q);
      break;
    }
      
    /* post the request */
    mx__post_ureq_null(ureq, &q->send.dest_address, partner, 
		       q->send.match_info, send_seq, mcp_handle);
    
    /* Small and medium messages are buffered immediately. */
    q->send.basic.state = MX__REQUEST_STATE_BUFFERED;
    q->send.basic.status.code = MX_STATUS_SUCCESS;
    q->send.basic.status.xfer_length = length;
    q->send.basic.status.msg_length = length;
    mx__enqueue_request(&ep->buffered_sendq, q);
    break;

  case MX__REQUEST_TYPE_SEND_SMALL:
    /* Resends already have a buffer allocated. This is pretty
       convoluted. We want to use the same sending code so below, we
       don't copy data if it's a requeue. */
    if (q->send.requeued) {
      p = ep->udataq + q->send.offset_data;
      shadow = ep->shadow_udataq + q->send.offset_data;
      
      /* In case of a requeue, always copy data from the shadowq back
	 into the udataq. This is a waste of time in the normal
	 requeue case but makes parity error recovery simpler. */
      mx_pio_memcpy(p, shadow, q->send.length);
      
      /* post the request */
      mx__post_ureq_small(ureq, &q->send.dest_address, partner, 
			  q->send.match_info, length, send_seq, mcp_handle, 
			  q->send.offset_data);
      mx__enqueue_request(&ep->buffered_sendq, q);
      break;
    }
    
    /* !q->send.requeued */
    p = mx__ptr_stack_pop(ep->small_msg_ptrs);
    if (p != NULL) {
      mx_assert(((uintptr_t)p % ep->small_msg_threshold) == 0);
      mx__copy_from_segments(p, q->send.segments, q->send.count,
			     q->send.memory_context, 0, length);
      q->send.offset_data = (uint32_t)(p - ep->udataq);
      
      /* post the request */
      mx__post_ureq_small(ureq, &q->send.dest_address, partner, 
			  q->send.match_info, length, send_seq, mcp_handle, 
			  q->send.offset_data);
      
      /* Small and medium messages are buffered immediately. */
      shadow = mx__ptr_stack_pop(ep->shadow_small_msg_ptrs);
      mx__copy_from_segments(shadow, q->send.segments, q->send.count,
			     q->send.memory_context, 0, length);
      q->send.basic.state = MX__REQUEST_STATE_BUFFERED;
      q->send.basic.status.code = MX_STATUS_SUCCESS;
      q->send.basic.status.xfer_length = length;
      mx__enqueue_request(&ep->buffered_sendq, q);
      break;
    }
    q->send.basic.type = MX__REQUEST_TYPE_SEND_MEDIUM;
    
  case MX__REQUEST_TYPE_SEND_MEDIUM:
    if (q->send.requeued) {
      p = ep->sendq + (q->send.offset_data << MX_MCP_VPAGE_SHIFT);
      i = (length + MX_MCP_VPAGE_SIZE - 1) >> MX_MCP_VPAGE_SHIFT;
      
      /* post the request */
      mx__post_ureq_medium(ureq, &q->send.dest_address, partner, 
			   q->send.match_info, length, send_seq, mcp_handle, 
			   q->send.offset_data, i,
			   ep->medium_msg_pipelines[length >> 10]);
      mx__enqueue_request(&ep->buffered_sendq, q);
      break;
    }
    
    /* !q->send.requeued */
    p = mx__memory_pool_alloc(ep->send_pool, ep->medium_msg_threshold);
    if (p != NULL) {
      uint32_t copy_len;
      mx_assert(((uintptr_t)p & MX_MCP_VPAGE_MASK) == 0);
      q->send.offset_data = (uint32_t) (p - ep->sendq) >> MX_MCP_VPAGE_SHIFT;
      
      copy_len = length <= MX_MCP_VPAGE_SIZE ? length : MX_MCP_VPAGE_SIZE;
      mx__copy_from_segments(p, q->send.segments, q->send.count,
			     q->send.memory_context, 0, copy_len);
      
      /* post the request */
      mx__post_ureq_medium(ureq, &q->send.dest_address, partner, 
			   q->send.match_info, length, send_seq, mcp_handle, 
			   q->send.offset_data, 1,
			   ep->medium_msg_pipelines[length >> 10]);
      q->send.basic.status.xfer_length = length;
      
      /* notify the MCP after copying each vpage of data into the sendq */
      if (length > MX_MCP_VPAGE_SIZE) {
	uint32_t copy_off = copy_len;
	length -= copy_len;
	p += copy_len;
	i = 2;
	do {
	  chunk = MX_MCP_VPAGE_SIZE;
	  if (length <= MX_MCP_VPAGE_SIZE) {
	    chunk = length;
	  }
	  
	  /* copy more data to the sendq */
	  mx__copy_from_segments(p, q->send.segments, q->send.count,
				 q->send.memory_context, copy_off, chunk);
	  
	  /* update the medium request */
	  ureq->medium.credits = htonl(i);
	  i++;
	  MX_STBAR();
	  
	  length -= chunk;
	  p += chunk;
	  copy_off += chunk;
	} while (length > 0);
      }
      
      /* Small and medium messages are buffered immediately. */
      q->send.basic.state = MX__REQUEST_STATE_BUFFERED;
      q->send.basic.status.code = MX_STATUS_SUCCESS;
      mx__enqueue_request(&ep->buffered_sendq, q);
      break;
    }
    mx_always_assert(0);
    
    
  case MX__REQUEST_TYPE_SEND_LARGE:
    if (!q->send.requeued) {
      /* FIXME: allocate ksend in clever way, for instance fix number of
	 segment in stack structure, otherwise malloc */
      mx_always_assert (q->send.count == 1);
      
      q->send.reg_area.nsegs = q->send.count;
      for (i = 0; i < q->send.count; ++i) {
	offset = (uint32_t)MX_RDMA_PAGE_OFFSET((uintptr_t)q->send.segments[i].segment_ptr);
	q->send.reg_area.segs[i].vaddr = ((uintptr_t) q->send.segments[i].segment_ptr 
				  - offset);
	q->send.reg_area.segs[i].len = q->send.segments[i].segment_length ? 
	  MX_RDMA_PAGE_ALIGN(q->send.segments[i].segment_length + offset) : 0;
      }
      q->send.reg_area.memory_context = q->send.memory_context;
      mx_rdmawin_get(ep, &q->send.reg_area, 1);
      q->send.local_rdma_id = q->send.reg_area.rdma_id;
      mx_always_assert (ep->rdma_requests[q->send.local_rdma_id].req == 0);
      ep->rdma_requests[q->send.local_rdma_id].req = q;
    }
    /* we can move this line before pinning pages to overlap pinning
       and scout sending */
    /* post the request */
    mx__post_ureq_rndv(ureq, &q->send.dest_address, partner, 
		       q->send.match_info, length, send_seq, mcp_handle, 
		       q->send.reg_area.rdma_id, 
		       ep->rdma_requests[q->send.reg_area.rdma_id].seqno, 
		       (uint16_t) MX_RDMA_PAGE_OFFSET
		       ((uintptr_t) q->send.segments[0].segment_ptr));
    q->send.basic.state &= ~MX__REQUEST_STATE_SEND_QUEUED;
    q->send.basic.state |= MX__REQUEST_STATE_MCP;
    mx__enqueue_request(&ep->large_sendq, q);
    break;
  default:
    mx_fatal("mx_post_send:unknown type");
  }
  if (partner->quadrant_count[MX__QUADRANT(partner->send_seq + MX__QUADRANT_ONE)]) {
    mx_printf("Sequence number (peer nic_id=%08"PRIx64") are rolling over in mx_lib: bad link/eventq full?\n",
	      partner->nic_id);
    mx_always_assert(0);
  }

}

static void
mx__process_requests(mx_endpoint_t ep, union mx_request *reqq,
		     unsigned int req_count)
{
  unsigned int i;
  union mx_request *req;
  mcp_ureq_t *ureq;
  uint16_t mcp_handle, offset;

  /* TODO: Magic number. */
  if (ep->event_count >= 32) {
    *ep->flow = htonl(ep->eventq_flow);
    ep->event_count = 0;
  }

  /* TODO: Magic number. */
  for (i = 0; i < req_count; ++i) {
    if (mx__isempty_request(reqq) || ep->rdmas.max == ep->rdmas.count) {
      break;
    }
    if (!mx__endpoint_avail_mcp_handle(ep)) {
      break;
    }
    req = mx__dequeue_request(reqq);
    mx_assert(req->basic.state == MX__REQUEST_STATE_SEND_QUEUED ||
	      (req->send.requeued && 
	       (req->basic.state == MX__REQUEST_STATE_DEAD ||
		req->basic.state == MX__REQUEST_STATE_BUFFERED)));

    if ((req->basic.type == MX__REQUEST_TYPE_SEND_LARGE) 
	&& (ep->rdmas.send_count >= ep->rdmas.max / 2)) {
      mx__putback_request(reqq, req);
      break;
    }

    mcp_handle = mx__endpoint_alloc_mcp_handle(ep);
    ureq = mx__get_ureq(ep->req_ring, mcp_handle);
    mx__hm_set(ep->handle_map, mcp_handle, req);
    
    switch (req->basic.type) {
    case MX__REQUEST_TYPE_SEND_NULL:
    case MX__REQUEST_TYPE_SEND_SMALL:
    case MX__REQUEST_TYPE_SEND_MEDIUM:
    case MX__REQUEST_TYPE_SEND_LARGE:
      mx__post_send(ep, req, ureq, mcp_handle);
      break;
      
    case MX__REQUEST_TYPE_RECV_LARGE:
      req->basic.mcp_handle = mcp_handle;
      req->recv.reg_area.nsegs = req->recv.count;
      mx_always_assert (req->recv.count == 1);
      offset = (uint16_t)MX_RDMA_PAGE_OFFSET((uintptr_t)req->recv.segments[0].segment_ptr);
      req->recv.reg_area.segs[0].vaddr = (uintptr_t)req->recv.segments[0].segment_ptr - offset;
      req->recv.reg_area.segs[0].len = req->basic.status.xfer_length ?
	MX_RDMA_PAGE_ALIGN(req->basic.status.xfer_length + offset) : 0;
      req->recv.reg_area.memory_context = req->recv.memory_context;
      mx_rdmawin_get(ep, &req->recv.reg_area, 0);
      req->recv.local_rdma_id = req->recv.reg_area.rdma_id;
      mx_assert (ep->rdma_requests[req->recv.reg_area.rdma_id].req == 0);
      ep->rdma_requests[req->recv.reg_area.rdma_id].req = req;
      mx__post_large_recv(ep, req, ureq, mcp_handle, 
			  req->recv.reg_area.rdma_id, 0, offset);
      mx__enqueue_request(&ep->large_getq, req);
      break;
    case MX__REQUEST_TYPE_CONNECT:
      req->basic.mcp_handle = mcp_handle;
      req->connect.basic.state &= ~MX__REQUEST_STATE_SEND_QUEUED;
      req->connect.basic.state |= MX__REQUEST_STATE_MCP;
      mx__post_ureq_connect(ureq, req, mcp_handle, 0);
      mx__enqueue_request(&ep->connectq, req);
      break;
    case MX__REQUEST_TYPE_CONNECT_REPLY:
      req->connect.basic.state = MX__REQUEST_STATE_MCP;
      mx__post_ureq_connect(ureq, req, mcp_handle, 1);
      mx__enqueue_request(&ep->connect_replyq, req);
      break;

    default:
      mx_printf("process_requests:Unknown type %d\n", req->basic.type);
      mx_always_assert(0);
    }
  }
}

void
mx__luigi(mx_endpoint_t ep)
{
  if (ep->in_callback) 
    return;
  mx__process_events(ep, 8); /* TODO: Magic number. */
  mx__process_requests(ep, &ep->send_reqq, 8); /* TODO: Magic number. */
#if MX_USE_SHMEM
  if (!mx__opt.disable_shmem)
    mx__shmem_luigi(ep);
#endif
#if MX_DEBUG
  mx__conservation_of_matter(ep);
#endif
}

static void
mx__parity_error_recover(struct mx_endpoint *ep)
{
  union mx_request *req;

  /* Things that need to be done to recover from a parity error.
     1) Process all pending events in the event queue.
     2) Re-open the endpoint?
     3) Re-submit all pending requests */
  mx__process_events(ep, -1);
  /* TODO: Re-open endpoint? */

  /* Reset some stuff. */
  mx__endpoint_init_eventq(ep);
  mx__endpoint_init_recvq(ep);
  mx__mcp_request_ring_init(ep->req_ring,
			    ep->sram+ep->desc.user_reqq_offset,
			    ep->desc.user_reqq_len);

  /* Queue up pending requests. Kind of like what happens when you get
     an MX_MCP_UEVT_ERROR. */
  /* TODO: Put request back in order? */
  while (!mx__isempty_request(&ep->buffered_sendq)) {
    req = mx__dequeue_request(&ep->buffered_sendq);
    req->send.requeued = 1;
    mx__putback_request(&ep->send_reqq, req);
    mx__endpoint_free_mcp_handle(ep, req->basic.mcp_handle);
  }
  while (!mx__isempty_request(&ep->large_sendq)) {
    req = mx__dequeue_request(&ep->large_sendq);
    req->send.basic.state |= MX__REQUEST_STATE_SEND_QUEUED;
    req->send.basic.state &= ~MX__REQUEST_STATE_MCP;
    req->send.requeued = 1;
    mx__putback_request(&ep->send_reqq, req);
    mx__endpoint_free_mcp_handle(ep, req->basic.mcp_handle);
  }
  while (!mx__isempty_request(&ep->large_getq)) {
    req = mx__dequeue_request(&ep->large_getq);
    req->basic.state = MX__REQUEST_STATE_SEND_QUEUED;
    mx_rdmawin_release(ep, req->recv.local_rdma_id, 0,
		       req->recv.reg_area.segs);
    ep->rdma_requests[req->recv.local_rdma_id].req = 0;
    mx__putback_request(&ep->send_reqq, req);
    mx__endpoint_free_mcp_handle(ep, req->basic.mcp_handle);
  }
  /* All the mcp handles should be accounted for by now. */
  /* Dont' bother to process requests, let it happen naturally. */
  /*printf("mx__recover_endpoint\n");*/
  mx__recover_endpoint(ep->handle);
}

MX__THREAD_RETURN_T
mx__progress(void *p)
{
  mx_wait_t x;
  struct mx_endpoint *ep;

  MX__THREAD_INIT(&ep->thread);

  /* wake-up every second, to check for connect requests
     allow to complete a mx_connect independently
     of receiver activity
  */
  x.timeout = 1000; /* XXXX */
  x.status = 0;
  x.mcp_wake_events = 0;
  ep = p;
  do {
    mx_return_t ret;
    MX__MUTEX_LOCK(&ep->lock);
    /* do the wake accounting from the previous iteration: assume the
       only one to generate wake_events bringing us here is the MCP
       (true because other kind of wake are associated with terminal
       conditions that won't cause the loop to be reentered) */
    mx_assert(ep->wake_pending >= x.mcp_wake_events);
    if (x.mcp_wake_events) {
      ep->wake_pending -= 1;
      mx__endpoint_free_mcp_handle(ep,ep->wake_mcp_handle);
    }
    
    mx__luigi(ep);
    if (ep->app_waiting) {
      mx__app_wake(ep->handle);
    }
    if (ep->waiters > 0 || ep->peek_waiters > 0 || ep->probe_waiters > 0) {
      if (!ep->app_waiting)
	mx__uwake_req(ep);
    }
    MX__MUTEX_UNLOCK(&ep->lock);
    ret = mx__wait(ep->handle, &x);
    if (ret != 0) {
      mx_fatal("mx__wait failed, check kernel logs for error messages");
    }
    MX__MUTEX_LOCK(&ep->lock);
#if MX_USE_SHMEM
    if (!mx__opt.disable_shmem)
      ep->shm->shmq->waiting = 0;
#endif
    switch(x.status) {
    case MX_WAIT_STATUS_GOOD:
      break;

    case MX_WAIT_PARITY_ERROR_DETECTED:
      /* do something here */
      x.status = 0;  /* driver does not set this unless there 
			is an error to report */
      /*printf("MX_WAIT_PARITY_ERROR_DETECTED\n");*/
      break;

    case MX_WAIT_PARITY_ERROR_CORRECTED:
      /* do something here */
      x.status = 0;  /* driver does not set this unless there 
			is an error to report */
      /*printf("MX_WAIT_PARITY_ERROR_CORRECTED\n");*/
      mx__parity_error_recover(ep);
      break;

    case MX_WAIT_PARITY_ERROR_UNCORRECTABLE:
      mx_fatal("Uncorrectable parity error received.");
      break;

    case MX_WAIT_ENDPT_ERROR:
      mx_fatal("MCP detected endpoint error.");
      break;

    default:
      {	
	char s[64];
	mx_sprintf(s, "Unknown error %d received", x.status);
	mx_fatal(s);
      }
      break;
    }
    MX__MUTEX_UNLOCK(&ep->lock);
  } while (!ep->cancelled);

  MX__THREAD_EXIT_WITH_NOTHING(&ep->thread);
}

#if 0
static void
mx__print_req(mcp_ureq_t *p)
{
  uint16_t dest_endpt;
  uint32_t dest_session;
  uint32_t offset_data;
  uint32_t dest_lookup_hint;
  uint32_t match_cst;
  uint16_t match_var0;
  uint16_t match_var1;
  uint32_t length;
  uint32_t id;
  uint32_t type;
  
  type = ntohl(p->basic.type);
  switch(type) {
  case MX_MCP_UREQ_SEND_PIO:
  case MX_MCP_UREQ_SEND_COPY:
    dest_endpt = ntohs(p->send.req.dest_endpt);
    dest_session = ntohl(p->send.req.dest_session);
    offset_data = ntohl(p->send.req.offset_data);
    dest_lookup_hint = ntohl(p->send.req.dest_lookup_hint);
    match_cst = ntohl(p->send.req.match_cst);
    match_var0 = ntohs(p->send.req.match_var0);
    match_var1 = ntohs(p->send.req.match_var1);
    length = ntohl(p->send.req.length);
    id = ntohl(p->send.req.id);
    break;
  default:
    mx_assert(0);
    break;
  }
}
#endif

void
mx__nic_id_to_macaddr(uint64_t nic_id, uint8_t macaddr[6])
{
  int i;
  for (i=5;i>=0;i--) {
    macaddr[i] = (uint8_t)(nic_id & 0xff);
    nic_id >>= 8;
  }
}

void
mx__macaddr_to_nic_id(uint8_t macaddr[6], uint64_t *nic_idp)
{
  int i;
  uint64_t nic_id = 0;
  for (i=0;i<6;i++) {
    nic_id <<= 8;
    nic_id += macaddr[i];
  }
  *nic_idp = nic_id;
}

MX_FUNC(const char *) mx__nic_id_to_str(char *str, uint64_t nic_id, uint32_t strlen)
{
  mx_snprintf(str, strlen, "%02x:%02x:%02x:%02x:%02x:%02x", 
	      (uint8_t)(nic_id >> 40),
	      (uint8_t)(nic_id >> 32),
	      (uint8_t)(nic_id >> 24),
	      (uint8_t)(nic_id >> 16),
	      (uint8_t)(nic_id >> 8),
	      (uint8_t)(nic_id >> 0));
  str[strlen - 1] = 0;
  return str;
}


void
mx__finish_buffered_requests(mx_endpoint_t ep)
{
  while (!mx__isempty_request(&ep->buffered_sendq)) {
    mx__luigi(ep);
  }
}


#define RDMAWIN_ENTRIES 8

void mx_rdmawin_init(mx_endpoint_t ep)
{
  int i;
  
#ifndef MX_KERNEL
  if (mx__opt.rcache && !mx__regcache_works()) {
    mx_printf("warning:regcache incompatible with malloc\n");
    mx__opt.rcache = 0;
  }
  if (mx__opt.verbose) {
    MX_WARN(("rdmawin_cache_code=(compiled=%s/possible=%s/activated=%s)\n",
	     MX_RDMAWIN_CACHE ? "yes" : "no",
	     mx__regcache_works() ? "yes" : "no",
	     mx__opt.rcache ? "yes" : "no"
	     ));
  }
#endif
  if (!mx__opt.rcache)
    return;
  TAILQ_INIT(&ep->rdmawin_pinned);
  TAILQ_INIT(&ep->rdmawin_free);
  ep->rdmawin_items = mx_calloc(RDMAWIN_ENTRIES, sizeof(*ep->rdmawin_items));
  /* FIXME: proper error handling */
  mx_always_assert(ep->rdmawin_items);
  for (i=0;i < RDMAWIN_ENTRIES;i++) {
    TAILQ_INSERT_HEAD(&ep->rdmawin_free, ep->rdmawin_items+i, list);
  }
}


void mx_rdmawin_finalize(mx_endpoint_t ep)
{
  struct mx__rdmawin *r;

  if (!mx__opt.rcache)
    return;
  TAILQ_FOREACH(r, &ep->rdmawin_pinned, list) {
    mx_unpin(r->ep, r->rdma_id, r->send);
  }
  TAILQ_INIT(&ep->rdmawin_pinned);
  mx_free(ep->rdmawin_items);
#if MX__EP_COUNTERS
  if (mx__opt.verbose) {
    MX_WARN(("hit:%u miss:%u  hit-kbytes:%lu miss-kbytes:%lu\n", 
	     ep->counters.cache_hit_cnt, ep->counters.cache_miss_cnt,
	     ep->counters.cache_hit_kbytes, ep->counters.cache_miss_kbytes));
  }
#endif
}


uint16_t mx_checksum(void *vptr, unsigned len)
{
  char *ptr = vptr;
  uint32_t sum = 0;
  /* RFC 1071 */
  while( len > 1 )  {
    /*  This is the inner loop */
    sum += * (uint16_t *)ptr;
    sum = (sum & 0xffff) + (sum >> 16);
    len -= 2;
    ptr += 2;
  }
  
  /*  Add left-over byte, if any */
  if( len > 0 )
    sum += * (uint8_t *) ptr;
  
  /*  Fold 32-bit sum to 16 bits */
  while (sum>>16)
    sum = (sum & 0xffff) + (sum >> 16);
  return ~sum;
}
