/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2005 by Myricom, Inc.  All rights reserved.                 *
 *************************************************************************/

#ifndef MX__LIB_TYPES_H
#define MX__LIB_TYPES_H

#include "mx_auto_config.h"
#  include "myriexpress.h"
#include "mx__thread.h" /* MX_EVENT_T, MX__MUTEX_T, MX__THREAD_T */
#include "mx_io.h" /* mx_reg_t, mx_get_copyblock_t */
#include "mcp_events.h" /* mcp_uevt_t */
#  include "mx_debug.h" /* mx_assert */
#  include "mx__shim.h" /* mx_calloc */
#include "mx__rdma_alloc.h" /* mx_rdma_pool_t */
#undef SLIST_ENTRY
#include "bsd/queue.h" /* TAILQ */
#include "mcp_requests.h" /* mcp_ureq_t */

typedef struct
{
  struct mx__partner *partner;
  uint16_t peer_index_n;
  uint8_t endpoint_id;
  uint8_t pad2[13 - sizeof(struct mx__partner *)];
} mx__endpoint_addr_t;

/* when changing one of the two enums below, please update mx__debug_dump.c
   for meaningful diagnostics */

enum mx__request_type {
  MX__REQUEST_TYPE_SEND_NULL = 0,
  MX__REQUEST_TYPE_SEND_SMALL,
  MX__REQUEST_TYPE_SEND_MEDIUM,
  MX__REQUEST_TYPE_SEND_LARGE,
  MX__REQUEST_TYPE_RECV,
  MX__REQUEST_TYPE_RECV_LARGE,
  MX__REQUEST_TYPE_RECV_SELF,
  MX__REQUEST_TYPE_RECV_SHM,
  MX__REQUEST_TYPE_CONNECT,
  MX__REQUEST_TYPE_CONNECT_REPLY
};

#if MX_ONE_SIDED
#define MX__ONESIDED_MASK ((uint64_t)3 << 62)
#define MX__ONESIDED_PUT ((uint64_t)1 << 62)
#define MX__ONESIDED_GET ((uint64_t)2 << 62)
#endif

enum mx__request_state {
  MX__REQUEST_STATE_SEND_QUEUED = 1, /* all buf non-large recv */
  MX__REQUEST_STATE_PENDING = 2, /* recv, unexp */
  MX__REQUEST_STATE_BUFFERED = 4, /* null/small/medium send */
  MX__REQUEST_STATE_COMPLETED = 8,  /* all */
  MX__REQUEST_STATE_DEAD = 16, /* null/small/medium send */
  MX__REQUEST_STATE_ACKED = 32, /* connect */
  MX__REQUEST_STATE_MCP = 64, /* large send/recv, connect */
  MX__REQUEST_STATE_NOTIFIED = 128, /* large send */
};

enum mx__wait_state {
  MX__WAIT_STATE_COMPLETE = 1,
  MX__WAIT_STATE_WAITING = 2,
  MX__WAIT_STATE_ABORTED = 3
};

struct mx__basic_request {
  enum mx__request_type type;
  enum mx__request_state state;
  union mx_request *prev;
  union mx_request *next;
  union mx_request *part_prev;
  union mx_request *part_next;
  mx_status_t status;
  uint16_t mcp_handle;

  enum mx__wait_state wait_state;
  MX__EVENT_T event;
};

typedef struct {
  uint64_t data_target;
  uint32_t length;
  uint32_t pad;
} mx__get_t;

union mx_request {
  struct {
    struct mx__basic_request basic;

    mx_segment_t *segments;
    uint32_t count;
    mx_segment_t segment;
    uintptr_t memory_context;
    mx__endpoint_addr_t dest_address;
    uint64_t match_info;
    uint32_t local_rdma_id;
    uint32_t offset_data;
    uint32_t length;
    mx_reg_t reg_area;

    int requeued; /* true if the lib is requeuing this request. Don't
		     use a new sequence number, use what the lib
		     provides. */
    uint16_t send_seq;
    struct mx__partner *partner;
    void *caller; /* debug: pointer to caller of mx_isend */
    uint64_t shm_magic;
  } send;
  struct {
    struct mx__basic_request basic;

    mx_segment_t *segments;
    uint32_t count;
    mx_segment_t segment; /* if only one, which is the general case */
    uintptr_t memory_context;
    uint64_t match_info;
    uint64_t match_mask;

    uint32_t offset_data;
    uint32_t r_length;
    uint16_t msg_seq;
    uint32_t accum; /* amount of bytes received but not necessarily copied */
    uint32_t r_mask; /* bitmask of pieces to receive */
    uint8_t unexpected;
    uint8_t put_target;
    uint8_t self_send;
    union mx_request *self_send_req;
    mx_reg_t reg_area;
    uint32_t local_rdma_id;
    /* the rdma window, seqnum and offset are copied to/from lanai without
       endianess convertion, they should be considered opaque */
    uint32_t remote_rdma;
    struct mx__partner *partner;
    uint64_t shm_peer_req;
    uint64_t shm_src_va;
    uint16_t shm_peer_endpt;
    uint32_t ordered_unexp_weight;
  } recv;
  struct {
    struct mx__basic_request basic;

    uint64_t match_info;
    uint64_t match_mask;
  } probe;
  struct {
    struct mx__basic_request basic;
  } peek;
  struct {
    struct mx__basic_request basic;
    uint8_t peer_endpoint_id;
    uint16_t peer_index_n;
    uint16_t lib_seqnum;
    uint8_t status_code_n;
    uint8_t connect_seqnum_n;
    uint32_t app_key_n;
    uint32_t dest_session_n;
    uint32_t timeout;
  } connect;
  struct mx__basic_request basic;
};

struct mx__request_lookaside
{
  union mx_request *r;
  union mx_request *free_req;
  int count;
  int alloc_count;
};

struct mx__connect {
  MX__THREAD_T thread;
  uint32_t key;
};

#define MX__EP_COUNTERS 1


/* TODO: Use mx_debug functionality? */
#if MX__EP_COUNTERS
struct mx__ep_counters
{
  uint32_t unexpected;
  uint32_t early;
  uint32_t no_mcp_req;
  uint32_t cache_hit_cnt;
  uint32_t cache_miss_cnt;
  unsigned long cache_hit_kbytes;
  unsigned long cache_miss_kbytes;
};
#define MX__EP_COUNTERS_INC(endpoint, field) endpoint->counters.field++
#define MX__EP_COUNTERS_ADD(endpoint, field, val) \
      do { endpoint->counters.field += (val); } while (0)
#else
#define MX__EP_COUNTERS_INC(endpoint, field)
#define MX__EP_COUNTERS_ADD(endpoint, field, val)
#endif

struct mx_endpoint
{
  MX__MUTEX_T lock;
  MX__THREAD_T thread;
  int cancelled;
  struct mx_endpoint *next;

  mx_endpt_handle_t handle;
  mx_get_copyblock_t desc;
  char *sram;

  char *sendq;
  struct mx__memory_pool *send_pool;

  /* EVENT: */
  char *eventq;
  int eventq_length;
  volatile mcp_uevt_t *eventq_uevt;
  uint32_t eventq_index;
  uint32_t eventq_flow;
  uint32_t event_count; /* number of events processed */

  /* RECV: */
  char *recvq;
  int recvq_length;
  int recvq_loc;

  struct mx__mcp_request_ring *req_ring;

  char *udataq;
  struct mx__ptr_stack *small_msg_ptrs;
  char *shadow_udataq;
  struct mx__ptr_stack *shadow_small_msg_ptrs;

  /* TODO: More fields may be added in the future in which case this
     becomes a struct. */
  uint32_t *flow;

  struct mx__request_lookaside req_lookaside;

  union mx_request send_reqq;
  union mx_request recv_reqq;
  union mx_request doneq;
  union mx_request unexpq;
  union mx_request buffered_sendq;
  union mx_request large_sendq;
  union mx_request large_getq;
  union mx_request multifrag_recvq;
  union mx_request probeq;
  union mx_request peekq;
  union mx_request connectq;
  union mx_request connect_replyq;
  uint32_t peek_count;
  uint32_t probe_count;
  uint32_t srdma_count;
  uint32_t sendself_count;
  uint32_t sendshm_count;

  struct mx__handle_map *handle_map;
  uint32_t max_mcp_handles;

  mx_rdma_pool_t rdmas;
  struct mx_rdma_req {
    union mx_request *req;
    uint8_t seqno;
#if MX_OS_UDRV
    void *copy;
    uintptr_t copy_offset;
#endif
  } *rdma_requests;

  mx__endpoint_addr_t addr;
  uint32_t endpoint_sid_n;
  int board_num;

  uint32_t small_msg_threshold;  /* threshold between pio and copy */
  uint32_t medium_msg_threshold; /* threshold between copy and dma */
  uint8_t medium_msg_pipelines[64];
  
  uint32_t max_endpoints;
  uint32_t min_light_endpoints;
  uint32_t max_light_endpoints;
  uint32_t max_peers;
  struct mx__partner **remote_ep;
  int32_t *application_rank_info; /* used for diagnostic only */
  struct mx__connect connect;
  struct mx__ep_counters counters;
  TAILQ_HEAD(s_rdmawin_pinned, mx__rdmawin) rdmawin_pinned;
  TAILQ_HEAD(s_rdmawin_free, mx__rdmawin) rdmawin_free;
  struct mx__rdmawin *rdmawin_items;
  uint32_t wake_pending;
  int wake_mcp_handle;
  int waiters;
  int peek_waiters;
  int probe_waiters;
  mx_matching_callback_t callback;
  void *callback_context;
  int in_callback;
  int app_waiting;					  

  mx_error_handler_t error_handler;
  uint32_t unexp_queue_length;
  uint32_t ordered_unexp_length;
  uint32_t unexp_queue_length_max;
  uint64_t get_id;

  struct mx__shm_info *shm;
};

typedef void (*mx__process_recv_t)(mx_endpoint_t ep, union mx_request *r, 
				   mcp_uevt_recv_t *evt, void *data);

struct mx__early
{
  struct mx__early *prev;
  struct mx__early *next;
  /* EVENT: */
  mcp_uevt_recv_inline_t recv_inline;
  uint8_t type;
  mx__process_recv_t recv_func;
  char *data;
  uint16_t data_length;
  uint16_t msg_seq;
};

struct mx__partner
{
  uint16_t send_seq;
  uint16_t recv_seq;
  uint8_t quadrant_count[4];
  union mx_request partialq;
  struct mx__early early;
  uint64_t nic_id;
  uint32_t endpoint_sid_n;
  uint32_t connect_session_n;
  int8_t connect_recvseq;
  int8_t connect_sendseq;
};

#endif
