/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2004 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

static const char __idstring[] = "@(#)$Id: mx__partner.c,v 1.24 2005/08/29 17:30:06 eugene Exp $";

#include "mx_auto_config.h"
#include "myriexpress.h"
#include "mx__lib_types.h"
#include "mx__partner.h"
#include "mx_byteswap.h"
#include "mx__driver_interface.h"
#include "mx__shmem.h"
#include "mx__lib.h"

/* TODO: Defer copy of event? */
struct mx__early *
mx__partner_insert_early(struct mx__partner *partner,
			 mcp_uevt_recv_t *recv,
			 uint16_t msg_seq,
			 mx__process_recv_t recv_func,
			 uint8_t type, char *data)
{
  struct mx__early *x;
  struct mx__early *y;
  uint32_t frame_length;

  for (x = partner->early.next;
       x != &partner->early;
       x = x->next) {
    if (mx__msg_order(x->msg_seq, msg_seq) > 0) {
      break;
    }
  }
  /* Either at end or x is first entry with larger sequence number.
     Insert before this one. */
  y = mx_malloc(sizeof (*y));
  if (y != NULL) {
    switch (type) {
    case MX_MCP_UEVT_RECV_NULL:
    case MX_MCP_UEVT_RECV_COPY_SINGLE:
      mx_memcpy(&(y->recv_inline), recv, sizeof (mcp_uevt_recv_copy_single_t));
      break;
    case MX_MCP_UEVT_RECV_INLINE:
    case MX_MCP_UEVT_RECV_RNDV:
      mx_memcpy(&(y->recv_inline), recv, sizeof (mcp_uevt_recv_inline_t));
      break;
    case MX_MCP_UEVT_RECV_COPY_MULTI:
      mx_memcpy(&(y->recv_inline), recv, sizeof (mcp_uevt_recv_copy_multi_t));
      break;
    default:
      mx_printf("inserting unknown early type %d\n", type);
      mx_fatal("bailing out");
      break;
    }
    /* EVENT: Copy recv buffer. */
    y->data = data;
    frame_length = ntohs(recv->length);
    if(type == MX_MCP_UEVT_RECV_COPY_MULTI) {
      mcp_uevt_recv_copy_multi_t *copy = (void*)recv;
      frame_length = ntohs(copy->frame_length);
    } 
    if (data) {
      y->data = mx_malloc(frame_length);
      if (y->data == NULL) {
	mx_free(y);
	return NULL;
      }
      mx_memcpy(y->data, data, frame_length);
    } else {
      mx_assert(data == NULL);
    }
    y->msg_seq = msg_seq;
    y->recv_func = recv_func;
    y->type = type;
    y->prev = x->prev;
    y->next = x;
    y->prev->next = y;
    y->next->prev = y;
  }
  return y;
}


void
mx__partner_cleanup(struct mx_endpoint *ep, struct mx__partner *p)
{
  p->recv_seq += MX__SESNO(p->recv_seq + MX__SESNO_ONE) | MX__SEQNO(p->recv_seq + 1000);
  if (p->partialq.basic.part_next != &p->partialq) {
    mx_printf("upon connect receive, partial requests for this peer..\n");
    /* TODO remove them */
  }
  if (p->early.next != &p->early) {
    mx_printf("upon connect receive, early requests for this peer..\n");
    /* TODO remove them */
  }
}

struct mx__partner *
mx__partner_create(struct mx_endpoint *ep, uint16_t peer_index, uint16_t endpt)
{
  struct mx__partner *p;
  mx_lookup_peer_t x;
  mx_return_t rc;
  p = mx_calloc(1, sizeof (struct mx__partner));
  mx_fixme_assert(p);
  p->partialq.basic.part_prev = &p->partialq;
  p->partialq.basic.part_next = &p->partialq;
  p->early.prev = &p->early;
  p->early.next = &p->early;
  p->endpoint_sid_n = -1;
  p->connect_session_n = -1;
  if (mx__opt.no_myrinet) {
    p->nic_id = UINT64_C(0x123456789abc);
  } else {
    x.index = peer_index;
    x.board_number = -1; /* not sure why it exists at all */
    rc = mx__peer_index_to_nic_id(ep->handle, &x);
    mx_fixme_assert(rc == MX_SUCCESS);
    p->nic_id = x.nic_id;
  }
  return p;
}
