/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2004 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#ifndef MX__REQUEST_H
# define MX__REQUEST_H

int mx__rl_init(struct mx__request_lookaside *rl);
void mx__rl_fini(struct mx__request_lookaside *rl);

uint16_t mx_checksum(void *ptr, unsigned length);

static inline int mx__count_requests(union mx_request *reqq)
{
  int count;
  union mx_request *p;

  for (count = 0, p = reqq->basic.next; p != reqq; ++count, p = p->basic.next);
  return count;
}

static inline void mx__init_request_list(union mx_request *list)
{
  list->basic.prev = list;
  list->basic.next = list;
}

static inline void mx__enqueue_request(union mx_request *reqq, union mx_request *req)
{
  mx_assert(req->basic.prev == NULL);
  mx_assert(req->basic.next == NULL);
  reqq->basic.prev->basic.next = req;
  req->basic.prev = reqq->basic.prev;
  req->basic.next = reqq;
  reqq->basic.prev = req;
}

static inline union mx_request *mx__dequeue_request(union mx_request *reqq)
{
  union mx_request *req;

  mx_assert(reqq->basic.next != reqq);
  req = reqq->basic.next;
  req->basic.prev->basic.next = req->basic.next;
  req->basic.next->basic.prev = req->basic.prev;
#if MX_DEBUG
  req->basic.prev = NULL;
  req->basic.next = NULL;
#endif
  return req;
}

static inline void mx__putback_request(union mx_request *reqq, union mx_request *req)
{
  mx_assert(req->basic.prev == NULL);
  mx_assert(req->basic.next == NULL);
  reqq->basic.next->basic.prev = req;
  req->basic.next = reqq->basic.next;
  req->basic.prev = reqq;
  reqq->basic.next = req;
}

static inline void mx__spliceout_request(union mx_request *list, union mx_request *req)
{
#if MX_DEBUG
  union mx_request *p;

  for (p = list->basic.next; p != list; p = p->basic.next) {
    if (p == req) {
      break;
    }
  }
  mx_always_assert(p == req);
#endif
  req->basic.prev->basic.next = req->basic.next;
  req->basic.next->basic.prev = req->basic.prev;
#if MX_DEBUG
  req->basic.prev = NULL;
  req->basic.next = NULL;
#endif
}

static inline int mx__isempty_request(union mx_request *reqq)
{
  return reqq->basic.next == reqq;
}

static inline union mx_request *mx__head_request(union mx_request *reqq)
{
  mx_assert(reqq->basic.next != reqq);
  return reqq->basic.next;
}

static inline
union mx_request *
mx__find_recv_request_by_match_info(union mx_request *anchor,
				    uint64_t match_info)
{
  union mx_request *r;

  for (r = anchor->basic.next; r != anchor; r = r->basic.next) {
    if (r->recv.match_info == (match_info & r->recv.match_mask)) {
      return r;
    }
  }
  return NULL;
}  

static inline
union mx_request *
mx__find_partial_recv_request_by_msg_seq(union mx_request *anchor,
					 uint64_t msg_seq)
{
  union mx_request *r;
  
  for (r = anchor->basic.part_next; r != anchor; r = r->basic.part_next) {
    if (msg_seq == r->recv.msg_seq) {
      return r;
    }
  }
  return NULL;
}

static inline
union mx_request *
mx__find_unex_recv_request_by_matching(union mx_request *anchor,
				       uint64_t match_info,
				       uint64_t match_mask)
{
  union mx_request *r;

  for (r = anchor->basic.next; r != anchor; r = r->basic.next) {
    /* The matching looks backwards compared to match_recv_request. */
    if ((r->recv.basic.status.match_info & match_mask) == match_info) {
      return r;
    }
  }
  return NULL;
}

static inline void mx__enqueue_part_request(union mx_request *reqq,
					    union mx_request *req)
{
  reqq->basic.part_prev->basic.part_next = req;
  req->basic.part_prev = reqq->basic.part_prev;
  req->basic.part_next = reqq;
  reqq->basic.part_prev = req;
}

static inline void mx__spliceout_part_request(union mx_request *list, union mx_request *req)
{
#if MX_DEBUG
  union mx_request *p;

  for (p = list->basic.part_next; p != list; p = p->basic.part_next) {
    if (p == req) {
      break;
    }
  }
  mx_always_assert(p == req);
#endif
  req->basic.part_prev->basic.part_next = req->basic.part_next;
  req->basic.part_next->basic.part_prev = req->basic.part_prev;
#if MX_DEBUG
  req->basic.part_prev = NULL;
  req->basic.part_next = NULL;
#endif
}

#ifndef MX_KERNEL
#define REQ_CHUNK 1024
#else
#define REQ_CHUNK (4000/sizeof(union mx_request) - 1)
#endif

static union mx_request *
mx__rl__alloc(void)
{
  union mx_request *r;
  int i;

  r = mx_malloc (REQ_CHUNK * sizeof (*r));
  if (r != NULL) {
    for (i = 1; i < REQ_CHUNK-1; ++i) {
      r[i].basic.next = &r[i+1];
    }
    r[REQ_CHUNK-1].basic.next = NULL;
    r[0].basic.next = NULL;
  }
  return r;
}


static inline union 
mx_request *mx__rl_alloc(struct mx__request_lookaside *rl)
{
  union mx_request *r;

  if (rl->count == 0) {
    r = mx__rl__alloc();
    if (r != NULL) {
      r[0].basic.next = rl->r;
      rl->r = r;
      rl->count = REQ_CHUNK-1;
      rl->alloc_count += REQ_CHUNK-1;
      rl->free_req = &rl->r[1];
    }
  }

  if (rl->count == 0) {
    return NULL;
  }
  else {
    r = rl->free_req;
    rl->free_req = rl->free_req->basic.next;
    rl->count--;
#if MX_DEBUG
    r->basic.next = NULL;
    r->basic.prev = NULL;
#endif
    return r;
  }
}
  
static inline void 
mx__rl_free(struct mx__request_lookaside *rl, union mx_request *r)
{
  mx_assert(r->basic.prev == NULL);
  mx_assert(r->basic.next == NULL);
  r->basic.next = rl->free_req;
  rl->free_req = r;
  rl->count++;
}

#endif /* REQUEST_H */
