/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2004 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#ifndef MX__REQUESTS_H
#define MX__REQUESTS_H

static inline void 
mx_copy32_inline(void *to, uint32_t *from32, unsigned size)
{
  register volatile uint32_t *to32;
  uint32_t tmp;
  int i;

  mx_assert ((size > 0)
	     && !(size & 3) && !((uintptr_t)to & 3) 
	     && !((uintptr_t)from32 & 3));

  to32 = (volatile uint32_t *) to;
  for (i = size / 4 - 1; i; i--)
    {
      *to32++ = *from32++;
    }
  tmp = *from32;
  MX_STBAR();
  *to32 = tmp;
  MX_STBAR();
}

static inline void 
mx_copy64_inline(void *to, uint64_t *from64, unsigned size)
{
  register volatile uint64_t *to64;
  uint64_t tmp;
  int i;

  mx_assert ((size > 0)
             && !(size & 7) && !((uintptr_t)to & 7) 
             && !((uintptr_t)from64 & 7));

  to64 = (volatile uint64_t *) to;
  for (i = size / 8 - 1; i; i--)
    {
      *to64++ = *from64++;
    }
  tmp = *from64;
  MX_STBAR();
  *to64 = tmp;
  MX_STBAR();
}


#if defined __GNUC__ && ((MX_CPU_x86 && MX_ENABLE_SSE2) || MX_CPU_x86_64) && !defined MX_KERNEL
#if __GNUC__ >= 3 && defined __SSE2__
#define MX_ASM_XMM_DEP ,"xmm0", "xmm1"
#else
#define MX_ASM_XMM_DEP
#endif
static inline void
mx_copy_sse2(void *to, uint64_t *from)
{
  if (mx__opt.wc_fence) {
    __asm__ __volatile__(
		       /* sse: 16byte writes */
		       "movdqu    (%1),%%xmm0\n\t"
		       "movdqu  16(%1),%%xmm1\n\t"
		       "movdqa  %%xmm0,   (%0)\n\t"
		       "movq  %%xmm1,  16(%0)\n\t"
		       "sfence\n\t"
		       "psrldq $8, %%xmm1 \n\t"
		       "movq  %%xmm1,  24(%0)\n\t"
		       "sfence\n\t"
		       ::"a"(to), "d"(from) : "memory" MX_ASM_XMM_DEP);
  } else {
    __asm__ __volatile__(
		       /* sse: 16byte writes */
		       "movdqu    (%1),%%xmm0\n\t"
		       "movdqu  16(%1),%%xmm1\n\t"
		       "sfence\n\t"
		       "movdqa  %%xmm0,   (%0)\n\t"
		       "movdqa  %%xmm1,  16(%0)\n\t"
       		       "sfence\n\t"
		       ::"a"(to), "d"(from) : "memory" MX_ASM_XMM_DEP);
  }
}
#endif


static inline void
mx_copy_ureq(void *to, void *from)
{
#if defined __GNUC__ && ((MX_CPU_x86 && MX_ENABLE_SSE2) || MX_CPU_x86_64) && !defined MX_KERNEL
  mx_copy_sse2(to, from);
#else
  if (sizeof (void *) == 8 || MX_OS_MACOSX == 1) {
    mx_copy64_inline(to, from, sizeof (mcp_ureq_t));
  } else {
    mx_copy32_inline(to, from, sizeof (mcp_ureq_t));
  }
#endif
}

static inline void
mx__post_ureq_null(mcp_ureq_t *req, mx__endpoint_addr_t *endpoint_addr,
		    struct mx__partner *partner, uint64_t match_info,
		    uint16_t seqnum, uint16_t cookie)
{
  mcp_ureq_t batch;
  uint32_t match_a, match_b;
  
  match_a = (uint32_t) (match_info >> 32);
  match_b = (uint32_t) (match_info & 0xffffffff);

  batch.null.dest_peer_index = endpoint_addr->peer_index_n;
  batch.null.dest_endpt = endpoint_addr->endpoint_id;
  batch.null.dest_session = partner->endpoint_sid_n;
  batch.null.match_a = htonl(match_a);
  batch.null.match_b = htonl(match_b);
  batch.null.lib_seqnum = htons(seqnum);
  batch.null.lib_cookie = htons(cookie);
  batch.null.type = MX_MCP_UREQ_SEND_NULL;

  mx_copy_ureq(req, batch.int_array);
}

static inline void
mx__post_ureq_small(mcp_ureq_t *req, mx__endpoint_addr_t *endpoint_addr,
		    struct mx__partner *partner, uint64_t match_info,
		    uint32_t length, uint16_t seqnum, uint16_t cookie, 
		    uint16_t offset)
{
  mcp_ureq_t batch;
  uint32_t match_a, match_b;
  
  match_a = (uint32_t) (match_info >> 32);
  match_b = (uint32_t) (match_info & 0xffffffff);

  batch.small.dest_peer_index = endpoint_addr->peer_index_n;
  batch.small.dest_endpt = endpoint_addr->endpoint_id;
  batch.small.dest_session = partner->endpoint_sid_n;
  batch.small.length = htonl(length);
  batch.small.match_a = htonl(match_a);
  batch.small.match_b = htonl(match_b);
  batch.small.lib_seqnum = htons(seqnum);
  batch.small.lib_cookie = htons(cookie);
  batch.small.offset = htons(offset);
  batch.small.type = MX_MCP_UREQ_SEND_SMALL;

  mx_copy_ureq(req, batch.int_array);
}

static inline void
mx__post_ureq_medium(mcp_ureq_t *req, mx__endpoint_addr_t *endpoint_addr,
		     struct mx__partner *partner, uint64_t match_info,
		     uint32_t length, uint16_t seqnum, uint16_t cookie, 
		     uint16_t sendq_index, uint32_t credits, uint8_t pipeline)
{
  mcp_ureq_t batch;
  uint32_t match_a, match_b;
  
  match_a = (uint32_t) (match_info >> 32);
  match_b = (uint32_t) (match_info & 0xffffffff);

  batch.medium.dest_peer_index = endpoint_addr->peer_index_n;
  batch.medium.dest_endpt = endpoint_addr->endpoint_id;
  batch.medium.dest_session = partner->endpoint_sid_n;
  batch.medium.length = htonl(length);
  batch.medium.match_a = htonl(match_a);
  batch.medium.match_b = htonl(match_b);
  batch.medium.lib_seqnum = htons(seqnum);
  batch.medium.lib_cookie = htons(cookie);
  batch.medium.sendq_index = htons(sendq_index);
  batch.medium.credits = htonl(credits);
  batch.medium.pipeline = pipeline;
  batch.medium.type = MX_MCP_UREQ_SEND_MEDIUM;

  mx_copy_ureq(req, batch.int_array);
}

static inline void
mx__post_ureq_rndv(mcp_ureq_t *req, mx__endpoint_addr_t *endpoint_addr,
		   struct mx__partner *partner, uint64_t match_info,
		   uint32_t length, uint16_t seqnum, uint16_t cookie, 
		   uint8_t rdma_id, uint8_t rdma_seqnum, uint16_t offset)
{
  mcp_ureq_t batch;
  uint32_t match_a, match_b;
  
  match_a = (uint32_t) (match_info >> 32);
  match_b = (uint32_t) (match_info & 0xffffffff);

  batch.rndv.dest_peer_index = endpoint_addr->peer_index_n;
  batch.rndv.dest_endpt = endpoint_addr->endpoint_id;
  batch.rndv.dest_session = partner->endpoint_sid_n;
  batch.rndv.length = htonl(length);
  batch.rndv.match_a = htonl(match_a);
  batch.rndv.match_b = htonl(match_b);
  batch.rndv.lib_seqnum = htons(seqnum);
  batch.rndv.lib_cookie = htons(cookie);
  batch.rndv.rdmawin_id = rdma_id;
  batch.rndv.rdmawin_seqnum = rdma_seqnum;
  batch.rndv.offset = htons(offset);
  batch.rndv.type = MX_MCP_UREQ_SEND_RNDV;

  mx_copy_ureq(req, batch.int_array);
}

static inline void
mx__post_ureq_pull(mcp_ureq_t *req, mx__endpoint_addr_t *endpoint_addr,
		   struct mx__partner *partner, union mx_request *r, 
		   uint8_t origin_rdma_id, uint8_t origin_rdma_seqnum, 
		   uint16_t origin_rdma_offset, uint16_t cookie)
{
  mcp_ureq_t batch;
  
  batch.pull.target_peer_index = endpoint_addr->peer_index_n;
  batch.pull.target_endpt = endpoint_addr->endpoint_id;
  batch.pull.target_session = partner->connect_session_n;
  batch.pull.length = htonl(r->recv.basic.status.xfer_length);
  batch.pull.origin_rdmawin_id = origin_rdma_id;
  batch.pull.origin_rdmawin_seqnum = origin_rdma_seqnum;
  batch.pull.origin_rdma_offset = htons(origin_rdma_offset);
  *((uint32_t *) &(batch.pull.target_rdmawin_id)) = r->recv.remote_rdma;
  batch.pull.origin_lib_cookie = htons(cookie);
  batch.pull.type = MX_MCP_UREQ_SEND_PULL;

  mx_copy_ureq(req, batch.int_array);
}

static inline void
mx__post_ureq_connect(mcp_ureq_t *req, union mx_request *r, uint16_t cookie, 
		      uint8_t is_reply)
{
  mcp_ureq_t batch;
  
  batch.connect.dest_peer_index = r->connect.peer_index_n;
  batch.connect.dest_endpt = r->connect.peer_endpoint_id;
  batch.connect.dest_session = r->connect.dest_session_n;
  batch.connect.timeout = htonl(r->connect.timeout);
  batch.connect.app_key = r->connect.app_key_n;
  batch.connect.is_reply = is_reply;
  batch.connect.connect_seqnum = r->connect.connect_seqnum_n;
  batch.connect.status_code = r->connect.status_code_n;
  batch.connect.lib_seqnum = htons(r->connect.lib_seqnum);
  batch.connect.lib_cookie = htons(cookie);
  batch.connect.type = MX_MCP_UREQ_SEND_CONNECT;

  mx_copy_ureq(req, batch.int_array);
}

static inline void 
mx__post_ureq_wake(mcp_ureq_t *req, uint32_t eventq_flow)
{
  mcp_ureq_t batch;

  batch.wake.eventq_flow = htonl(eventq_flow);
  batch.wake.type = MX_MCP_UREQ_POST_WAKE;
  
  mx_copy_ureq(req, batch.int_array);
}


static inline void
mx__wake_request(struct mx_endpoint *ep, union mx_request *r)
{
  if (r->basic.wait_state == MX__WAIT_STATE_WAITING) {
    r->basic.wait_state = MX__WAIT_STATE_COMPLETE;
    ep->waiters--;
    MX__EVENT_SIGNAL(&r->basic.event);
  }
}


#endif
