/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2004 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#include "mx_auto_config.h"
#include "myriexpress.h"
#include "mx__lib_types.h"
#include "mx_util.h"
#include "mx__segment.h"
#include "mx__lib.h"
#  include "mx__partner.h"
#include "mx__endpoint.h"
#include "mx__request.h"

void
mx__self_copy(struct mx_endpoint *ep, union mx_request *sreq, union mx_request *rreq, void *unex_data)
{
  void *data;
  uint32_t xfer_len;

  data = sreq ? (void*)(uintptr_t)sreq->send.segments[0].segment_ptr :
    unex_data;

  xfer_len = MX_MIN(rreq->recv.r_length, rreq->recv.basic.status.msg_length);

  mx__copy_to_segments(rreq->recv.segments,
		       rreq->recv.count,
		       rreq->recv.memory_context,
		       0,		       
		       data,
		       xfer_len);

  memcpy(&rreq->recv.basic.status.source, &ep->addr,
	 sizeof (rreq->recv.basic.status.source));

  rreq->recv.basic.status.xfer_length = xfer_len;
  mx__recv_complete(ep, rreq, MX_STATUS_SUCCESS);

  if (sreq) {
    sreq->send.basic.status.xfer_length = xfer_len;
    mx__send_complete(ep, sreq, MX_STATUS_SUCCESS);
  }
}


void mx__self_send(struct mx_endpoint *ep, union mx_request *q)
{
  union mx_request *rreq;
  if (q->send.count > 1) {
    mx_fatal("too many send segments!!!");
  }
  q->send.basic.state = MX__REQUEST_STATE_PENDING;
  rreq = mx__endpoint_match_recv_request(ep, q->send.match_info, q->send.length);
  if (rreq) {
    mx__spliceout_request(&ep->recv_reqq, rreq);
    rreq->recv.basic.status.match_info = q->send.match_info;
    rreq->recv.basic.status.msg_length = q->send.length;
    mx__self_copy(ep, q, rreq, 0);
  } else {
    rreq = mx__rl_alloc(&ep->req_lookaside);
    if (rreq == NULL) {
      mx_fatal("mx__self_send:out of resources");
    }
    rreq->recv.partner = ep->addr.partner;
    rreq->recv.segments = &rreq->recv.segment;
    rreq->recv.count = 1;
    rreq->recv.segment.segment_length = 0;
    rreq->recv.segment.segment_ptr = 0;
    rreq->recv.basic.status.match_info = q->send.match_info;
    rreq->recv.basic.status.msg_length = q->send.length;
    rreq->recv.basic.type = MX__REQUEST_TYPE_RECV_SELF;
    rreq->recv.unexpected = 1;
    mx__enqueue_request(&ep->unexpq, rreq); 
    if (q->send.basic.type == MX__REQUEST_TYPE_SEND_LARGE) {
      rreq->recv.basic.status.code = MX_STATUS_PENDING;
      rreq->recv.basic.state = MX__REQUEST_STATE_PENDING;
      rreq->recv.self_send_req = q;
      ep->sendself_count += 1;
    } else {
      rreq->recv.basic.status.code = MX_STATUS_SUCCESS;
      rreq->recv.basic.state = MX__REQUEST_STATE_COMPLETED;
      rreq->recv.self_send_req = 0;
      if (q->send.length) {
	void *segment_ptr = mx_malloc(q->send.length);
	if (segment_ptr == NULL) {
	  mx_fatal("Warning: mx__self_send/unexp:mx_malloc failed");
	}
	rreq->recv.segment.segment_ptr = MX_VA_TO_SEGMENT_PTR(segment_ptr);
	rreq->recv.segment.segment_length = q->send.length;
	mx__copy_from_segments(segment_ptr, q->send.segments, q->send.count,
			       q->send.memory_context, 0, q->send.length);
      }
      rreq->recv.ordered_unexp_weight = 0;
      ep->unexp_queue_length += q->send.length;
      
      q->send.basic.status.xfer_length = q->send.length;
      mx__send_complete(ep, q, MX_STATUS_SUCCESS);
    }
  }
}
