/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2004 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

static const char __idstring[] = "@(#)$Id: mx_isend.c,v 1.73 2005/11/01 16:26:27 loic Exp $";
 
#include "mx_auto_config.h"
#include "myriexpress.h"
#include "mx__lib_types.h"
#include "mx_pin.h"
#include "mx__request.h"
#include "mx__segment.h"
#include "mx__lib.h"
#include "mx__self.h"
#  include "mx__partner.h"
#include "mx__endpoint.h"
#include "mx__handle_map.h"
#include "mx__mcp_request_ring.h"
#include "mx__error.h"
#include "mx__shmem.h"

MX_FUNC(mx_return_t)
#ifndef MX_KERNEL
mx_isend(struct mx_endpoint * ep,
	 mx_segment_t *segments_list, uint32_t segments_count,
	 mx_endpoint_addr_t dest_address, uint64_t match_info,
	 void *context, mx_request_t *request)
#else
mx_kisend(struct mx_endpoint * ep,
	 mx_segment_t *segments_list, uint32_t segments_count,
	 mx_pin_type_t pin_type,
	 mx_endpoint_addr_t dest_address, uint64_t match_info,
	 void *context, mx_request_t *request)
#endif
{
  union mx_request *q;
  mcp_ureq_t *ureq;
  mx__endpoint_addr_t *dest;
  uint16_t mcp_handle;
  uint32_t length;
  mx_return_t ret_val;
#ifndef MX_KERNEL
  mx_pin_type_t pin_type = MX_PIN_UNDEFINED;
#endif

  ret_val = MX_SUCCESS;
  dest = (mx__endpoint_addr_t *)(void *)&dest_address;

  MX__MUTEX_LOCK(&ep->lock);
  q = mx__rl_alloc(&ep->req_lookaside);
  if (q == NULL) {
    ret_val = MX_NO_RESOURCES;
    goto abort_with_lock;
  }

  MX__CACHE_SEGMENTS(q->send);
  if ((q->send.count > 1) && (q->send.segments == NULL)) {
    ret_val = MX_NO_RESOURCES;
    goto abort_with_request;
  }
#ifndef MX_KERNEL
  if (MX_DEBUG && mx__opt.csum) {
    mx_always_assert(match_info >> 48 == 0);
    if (q->send.count)
      match_info |= 
	(uint64_t)mx_checksum(q->send.segments[0].segment_ptr, 
			      q->send.segments[0].segment_length) << 48;
  }
#endif
#if defined __GNUC__ && MX_DEBUG
  q->send.caller = __builtin_return_address(0);
#endif
  q->send.memory_context = pin_type;
  q->send.dest_address = *dest;
  q->send.match_info = match_info;
  q->send.basic.status.context = context;
  q->send.basic.mcp_handle = -1;
  q->send.basic.wait_state = MX__WAIT_STATE_COMPLETE;
  q->send.basic.state = MX__REQUEST_STATE_SEND_QUEUED;
  length = mx__total_segment_length(q->send.segments, q->send.count);
  q->send.length = length;
  if (length == 0) {
    q->send.basic.type = MX__REQUEST_TYPE_SEND_NULL;
  } else if (length <= ep->small_msg_threshold) {
    q->send.basic.type = MX__REQUEST_TYPE_SEND_SMALL;
  } else if (length <= ep->medium_msg_threshold) {
    q->send.basic.type = MX__REQUEST_TYPE_SEND_MEDIUM;
  } else {
    q->send.basic.type = MX__REQUEST_TYPE_SEND_LARGE;
  }
  q->send.requeued = 0;

#ifndef MX_KERNEL
  if (dest->peer_index_n == ep->addr.peer_index_n
      && dest->endpoint_id == ep->addr.endpoint_id
      && !mx__opt.disable_self) { 
    mx__self_send(ep, q);
  } else
#if MX_USE_SHMEM
    if (dest->peer_index_n == ep->addr.peer_index_n &&
	!mx__opt.disable_shmem) {
    mx__shm_send(ep,q);
  } else
#endif
#endif
    if (mx__isempty_request(&ep->send_reqq) &&
	     mx__endpoint_avail_mcp_handle(ep) &&
	     q->send.basic.type != MX__REQUEST_TYPE_SEND_LARGE) {
    mcp_handle = mx__endpoint_alloc_mcp_handle(ep);
    ureq = mx__get_ureq(ep->req_ring, mcp_handle);
    mx__hm_set(ep->handle_map, mcp_handle, q);
    mx__post_send(ep, q, ureq, mcp_handle);
  } else {
    mx__enqueue_request(&ep->send_reqq, q);
  }

  mx__luigi(ep);

  *request = q;
  MX__MUTEX_UNLOCK(&ep->lock);
  return MX_SUCCESS;

 abort_with_request:
  mx__rl_free(&ep->req_lookaside, q);
 abort_with_lock:
  MX__MUTEX_UNLOCK(&ep->lock);
  return mx__error(ep, "mx_isend", ret_val);
}
