/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2004 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

static const char __idstring[] = "@(#)$Id: mx_probe.c,v 1.28 2005/08/08 22:49:39 eugene Exp $";

#include "mx_auto_config.h"
#include "myriexpress.h"
#include "mx__lib_types.h"
#include "mx__request.h"
#include "mx__sleep.h"
#include "mx__error.h"

int mx__iprobe_common(mx_endpoint_t ep, uint64_t match_info,
		      uint64_t match_mask, mx_status_t *status);

MX_FUNC(mx_return_t)
mx_probe(mx_endpoint_t ep, uint32_t timeout, uint64_t match_info,
	 uint64_t match_mask, mx_status_t *status, uint32_t *result)
{
  union mx_request *r = NULL;
  union mx_request *probe_request;

  /* TODO: This whole thing is pretty convoluted. Find a cleaner way to
     express this */
  MX__MUTEX_LOCK(&ep->lock);
  /* Go through one iteration of mx__iprobe_common to see if a matching
     unexpected message is already here. */
  *result = mx__iprobe_common(ep, match_info, match_mask, status);
  if (*result == 0) {
    /* Not done, we want to post a uwake but there might not be an mcp
       request available. */
    probe_request = mx__rl_alloc(&ep->req_lookaside);
    if (probe_request == NULL) {
      MX__MUTEX_UNLOCK(&ep->lock);
      return mx__error(ep, "mx_probe", MX_NO_RESOURCES);
    }
    ep->probe_count += 1;
    probe_request->probe.match_info = match_info;
    probe_request->probe.match_mask = match_mask;
    probe_request->basic.status.context = NULL;
    mx__sleep_on_request(ep, probe_request, timeout, &ep->probe_waiters, &ep->probeq);
    r = probe_request->basic.status.context;
    if (r != NULL) {
      /* TODO: The is the same work done in mx_iprobe_common. */
      mx_memcpy(status, &r->recv.basic.status, sizeof (*status));
      *result = 1;
    }
    else {
      *result = 0;
    }
    ep->probe_count -= 1;
    mx__rl_free(&ep->req_lookaside, probe_request);
  }
  MX__MUTEX_UNLOCK(&ep->lock);
  return MX_SUCCESS;
}
