/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2004 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#include "mx_auto_config.h"
#include "myriexpress.h"
#include "mx__lib_types.h"
#include "mx__request.h"

#if MX_THREAD_SAFE
static void
mx__wake_queue(struct mx_endpoint *ep, union mx_request *list)
{
  union mx_request *req;
  for (req = list->basic.next;
       req != list;
       req = req->basic.next) {
    if (req->basic.wait_state == MX__WAIT_STATE_WAITING) {
      req->basic.wait_state = MX__WAIT_STATE_ABORTED;
      ep->waiters--;
      MX__EVENT_SIGNAL(&req->basic.event);
    }
  }
}
#endif

MX_FUNC(mx_return_t)
mx_wakeup(mx_endpoint_t ep)
{
#if MX_THREAD_SAFE
  union mx_request *req;

  MX__MUTEX_LOCK(&ep->lock);

  while (ep->peek_waiters > 0) {
    req = mx__dequeue_request(&ep->peekq);
    req->basic.wait_state = MX__WAIT_STATE_ABORTED;
    ep->peek_waiters--;
    MX__EVENT_SIGNAL(&req->basic.event);
  }

  while (ep->probe_waiters > 0) {
    req = mx__dequeue_request(&ep->probeq);
    req->basic.wait_state = MX__WAIT_STATE_ABORTED;
    ep->probe_waiters--;
    MX__EVENT_SIGNAL(&req->basic.event);
  }

  mx__wake_queue(ep, &ep->send_reqq);
  mx__wake_queue(ep, &ep->recv_reqq);
  mx__wake_queue(ep, &ep->buffered_sendq);
  mx__wake_queue(ep, &ep->large_sendq);
  mx__wake_queue(ep, &ep->large_getq);
  mx__wake_queue(ep, &ep->multifrag_recvq);

  MX__MUTEX_UNLOCK(&ep->lock);
  return MX_SUCCESS;
#else
  return MX_SUCCESS;
#endif
}
