/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2004 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

/* This file was originally contributed by
 * Brice.Goglin@ens-lyon.org (LIP/INRIA/ENS-Lyon) */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <semaphore.h>
#include <pthread.h>
#include <unistd.h>
#include "mx_auto_config.h"
#include "myriexpress.h"
#include "test_common.h"

mx_endpoint_t ep;
mx_endpoint_addr_t addr;
sem_t launched, stopped;

void * send1_func(void *dummy)
{
  mx_return_t ret;
  mx_status_t status;
  uint32_t result;
  mx_request_t req;
  mx_segment_t seg = {NULL, 0};
  
  ret = mx_issend(ep, &seg, 1, addr, 0, NULL, &req);
  insist(ret == MX_SUCCESS);

  sem_post(&launched);
  printf("send1 launched\n");

  ret = mx_wait(ep, &req, MX_INFINITE, &status, &result);
  insist(ret == MX_SUCCESS && !result);
  printf("send1 woke up\n");

  sem_post(&stopped);
  printf("send1 stopped\n");
  
  pthread_exit(NULL);
  return NULL;
}

void * send2_func(void *dummy)
{
  mx_return_t ret;
  mx_status_t status;
  uint32_t result;
  mx_request_t req;
  mx_segment_t seg = {NULL, 0};
  
  ret = mx_issend(ep, &seg, 1, addr, 0, NULL, &req);
  insist(ret == MX_SUCCESS);

  sem_post(&launched);
  printf("send2 launched\n");

  ret = mx_wait(ep, &req, MX_INFINITE, &status, &result);
  insist(ret == MX_SUCCESS && !result);
  printf("send2 woke up\n");

  sem_post(&stopped);
  printf("send2 stopped\n");
  
  pthread_exit(NULL);
  return NULL;
}

void * recv1_func(void *dummy)
{
  mx_return_t ret;
  mx_status_t status;
  uint32_t result;
  mx_request_t req;
  mx_segment_t seg = {NULL, 0};
  
  ret = mx_irecv(ep, &seg, 1, 1, MX_MATCH_MASK_NONE, NULL, &req);
  insist(ret == MX_SUCCESS);

  sem_post(&launched);
  printf("recv1 launched\n");

  mx_wait(ep, &req, MX_INFINITE, &status, &result);
  insist(ret == MX_SUCCESS && !result);
  ret = printf("recv1 woke up\n");

  sem_post(&stopped);
  printf("recv1 stopped\n");
  
  pthread_exit(NULL);
  return NULL;
}

void * recv2_func(void *dummy)
{
  mx_return_t ret;
  mx_status_t status;
  uint32_t result;
  mx_request_t req;
  mx_segment_t seg = {NULL, 0};
  
  ret = mx_irecv(ep, &seg, 1, 1, MX_MATCH_MASK_NONE, NULL, &req);
  insist(ret == MX_SUCCESS);

  sem_post(&launched);
  printf("recv2 launched\n");

  mx_wait(ep, &req, MX_INFINITE, &status, &result);
  insist(ret == MX_SUCCESS && !result);
  ret = printf("recv2 woke up\n");

  sem_post(&stopped);
  printf("recv2 stopped\n");
  
  pthread_exit(NULL);
  return NULL;
}

void * probe1_func(void *dummy)
{
  mx_return_t ret;
  mx_status_t status;
  uint32_t result;
  
  sem_post(&launched);
  printf("probe1 launched\n");

  ret = mx_probe(ep, MX_INFINITE, 2, 2, &status, &result);
  insist(ret == MX_SUCCESS && !result);
  printf("probe1 woke up\n");

  sem_post(&stopped);
  printf("probe1 stopped\n");
  
  pthread_exit(NULL);
  return NULL;
}

void * probe2_func(void *dummy)
{
  mx_return_t ret;
  mx_status_t status;
  uint32_t result;
  
  sem_post(&launched);
  printf("probe2 launched\n");

  ret = mx_probe(ep, MX_INFINITE, 2, 2, &status, &result);
  insist(ret == MX_SUCCESS && !result);
  printf("probe2 woke up\n");

  sem_post(&stopped);
  printf("probe2 stopped\n");
  
  pthread_exit(NULL);
  return NULL;
}

void * peek1_func(void *dummy)
{
  mx_return_t ret;
  uint32_t result;
  mx_request_t any;
    
  sem_post(&launched);
  printf("peek1 launched\n");

  ret = mx_peek(ep, MX_INFINITE, &any, &result);
  insist(ret == MX_SUCCESS && !result);
  printf("peek1 woke up\n");

  sem_post(&stopped);
  printf("peek1 stopped\n");
  
  pthread_exit(NULL);
  return NULL;
}

void * peek2_func(void *dummy)
{
  mx_return_t ret;
  uint32_t result;
  mx_request_t any;
  
  sem_post(&launched);
  printf("peek2 launched\n");

  ret = mx_peek(ep, MX_INFINITE, &any, &result);
  insist(ret == MX_SUCCESS && !result);
  printf("peek2 woke up\n");

  sem_post(&stopped);
  printf("peek2 stopped\n");
  
  pthread_exit(NULL);
  return NULL;
}

int main() {
  mx_return_t ret;
  pthread_t send1_th;
  pthread_t send2_th;
  pthread_t recv1_th;
  pthread_t recv2_th;
  pthread_t probe1_th;
  pthread_t probe2_th;
  pthread_t peek1_th;
  pthread_t peek2_th;
  int i;
  int n = 0;

  sem_init(&launched, 0, 0);
  sem_init(&stopped, 0, 0);

  ret = mx_init();
  insist(ret == MX_SUCCESS);
  ret = mx_open_endpoint(MX_ANY_NIC, MX_ANY_ENDPOINT, 0, NULL, 0, &ep);
  insist(ret == MX_SUCCESS);
  ret = mx_get_endpoint_addr(ep, &addr);
  insist(ret == MX_SUCCESS);
  
  printf("launching\n");
  pthread_create(&send1_th, NULL, &send1_func, NULL); n++;
  pthread_create(&send2_th, NULL, &send2_func, NULL); n++;
  pthread_create(&recv1_th, NULL, &recv1_func, NULL); n++;
  pthread_create(&recv2_th, NULL, &recv2_func, NULL); n++;
  pthread_create(&probe1_th, NULL, &probe1_func, NULL); n++;
  pthread_create(&probe2_th, NULL, &probe2_func, NULL); n++;
  pthread_create(&peek1_th, NULL, &peek1_func, NULL); n++;
  pthread_create(&peek2_th, NULL, &peek2_func, NULL); n++;

  for(i=0; i<n; i++)
    sem_wait(&launched);
  sleep(5);

  mx_wakeup(ep);
  for(i=0; i<n; i++)
    sem_wait(&stopped);

  printf("joining\n");
  pthread_join(send1_th, NULL);
  pthread_join(send2_th, NULL);
  pthread_join(recv1_th, NULL);
  pthread_join(recv2_th, NULL);
  pthread_join(peek1_th, NULL);
  pthread_join(peek2_th, NULL);
  pthread_join(probe1_th, NULL);
  pthread_join(probe2_th, NULL);

  ret = mx_close_endpoint(ep);
  insist(ret == MX_SUCCESS);
  ret = mx_finalize();
  insist(ret == MX_SUCCESS);

  return 0;
}
